import xtrack as xt
import numpy as np

env = xt.Environment()
env.particle_ref = xt.Particles(p0c=2e9, mass0=xt.PROTON_MASS_EV)

env['pi'] = np.pi
env['l_bend'] = 3.5
env['l_quad'] = 1.
env['l_cell'] = 20.
env['n_bends'] = 24.

env['h_bend']= 'pi / n_bends / l_bend'

env.new('mq', xt.Quadrupole, length='l_quad')
env.new('mb', xt.Bend, length='l_bend', h='h_bend', k0='h_bend')

env.new('mqf', 'mq', k1=0.1)
env.new('mqd', 'mq', k1=-0.1)

# We create a line which defines half of a FODO cell
arc_half_cell = env.new_line(components=[
    env.place('mqf'),
    env.new('mb1', 'mb', at='l_cell/4 - (l_bend/2 + 0.2)', from_='mqf'),
    env.new('mb2', 'mb', at='l_cell/4 + (l_bend/2 + 0.2)', from_='mqf'),
    env.place('mqd', at='l_cell/2 - l_quad/2', from_='mqf'),
    ])


# We create two replicas of the half cell, one mirrored and one not and we
# call them left and right
arc_half_cell_left = env.new('l', arc_half_cell, mirror=True, mode='replica')
arc_half_cell_right = env.new('r', arc_half_cell, mode='replica')

# We concatenate the two half cells to create a full cell
cell = arc_half_cell_left + arc_half_cell_right
cell.get_table()
# this gives the following (note that the name given to the line is added to the
# element name ad suffix):
#
# name                   s element_type isthick isreplica parent_name ...
# mqd.l                  0 Quadrupole      True      True mqd
# drift_3.l              1 Drift           True      True drift_3
# mb2.l                1.3 Bend            True      True mb2
# drift_2.l            4.8 Drift           True      True drift_2
# mb1.l                5.2 Bend            True      True mb1
# drift_1.l            8.7 Drift           True      True drift_1
# mqf.l                9.5 Quadrupole      True      True mqf
# mqf.r               10.5 Quadrupole      True      True mqf
# drift_1.r           11.5 Drift           True      True drift_1
# mb1.r               12.3 Bend            True      True mb1
# drift_2.r           15.8 Drift           True      True drift_2
# mb2.r               16.2 Bend            True      True mb2
# drift_3.r           19.7 Drift           True      True drift_3
# mqd.r                 20 Quadrupole      True      True mqd
# _end_point            21                False     False None

# Similarly, we can create two cells with different names
env.new('cell1', cell, mode='replica')
env.new('cell2', cell, mode='replica')

# And we can create an arc concatenating the two cells
env['arc'] = env['cell1'] + env['cell2']

# We can make 6 arcs with different names
env.new('arc1', 'arc', mode='replica')
env.new('arc2', 'arc', mode='replica')
env.new('arc3', 'arc', mode='replica')
env.new('arc4', 'arc', mode='replica')
env.new('arc5', 'arc', mode='replica')
env.new('arc6', 'arc', mode='replica')
# Also in this case the name of the line is added as a suffix to the element
# names.

# We concatenated them in a ring. This can be done using the '+' operator
# as shown before or using the arcs as subsections of a new line, i.e.:
ring = env.new_line(components=['arc1', 'arc2', 'arc3', 'arc4', 'arc5', 'arc6'])

# One can see that having used named lines and replicas to build the ring has
# generated a regular name pattern for the elements in the ring, as we can see
# for example by listing all focusing quadrupoles in the ring:
ring.get_table().rows['mqf.*']
# is:
#
# Table: 24 rows, 8 cols
# name                         s element_type isthick isreplica ...
# mqf.l.cell1.arc1           9.5 Quadrupole      True      True
# mqf.r.cell1.arc1          10.5 Quadrupole      True      True
# mqf.l.cell2.arc1          30.5 Quadrupole      True      True
# mqf.r.cell2.arc1          31.5 Quadrupole      True      True
# mqf.l.cell1.arc2          51.5 Quadrupole      True      True
# mqf.r.cell1.arc2          52.5 Quadrupole      True      True
# mqf.l.cell2.arc2          72.5 Quadrupole      True      True
# mqf.r.cell2.arc2          73.5 Quadrupole      True      True
# mqf.l.cell1.arc3          93.5 Quadrupole      True      True
# mqf.r.cell1.arc3          94.5 Quadrupole      True      True
# mqf.l.cell2.arc3         114.5 Quadrupole      True      True
# mqf.r.cell2.arc3         115.5 Quadrupole      True      True
# etc...


import xtrack as xt
import numpy as np

slice_mode = 'thin'

ele_test = [
    xt.Bend(k0=0.04, h=0.03, length=1,
            k1=0.1,
            knl=[0.7, 0.8, 0.9, 1.0], ksl=[0.1, 0.2, 0.3, 0.4],
            edge_entry_angle=0.05, edge_exit_angle=0.06,
            edge_entry_hgap=0.06, edge_exit_hgap=0.07,
            edge_entry_fint=0.08, edge_exit_fint=0.09,
            ),
    xt.Quadrupole(k1=2., k1s=-3., length=3.),
    xt.Sextupole(k2=0.1, k2s=0.2, length=0.3),
    xt.Octupole(k3=0.1, k3s=0.2, length=0.4),
    xt.Multipole(knl=[0.7, 0.8, 0.9, 1.0], ksl=[0.1, 0.2, 0.3, 0.4],
                        length=0.4, hxl=0.1)
]

for elem in ele_test:
    print('ele type:', elem.__class__.__name__)

    shift_x = 1e-3
    shift_y = 2e-3
    rot_s_rad = -0.4

    line_test = xt.Line(elements=[elem.copy()])

    line_ref = xt.Line(elements=[
        xt.XYShift(dx=shift_x, dy=shift_y),
        xt.SRotation(angle=np.rad2deg(rot_s_rad)),
        elem.copy(),
        xt.SRotation(angle=np.rad2deg(-rot_s_rad)),
        xt.XYShift(dx=-shift_x, dy=-shift_y),
    ])

    if slice_mode is not None:
        line_test.slice_thick_elements(
            slicing_strategies=[xt.Strategy(xt.Teapot(3, mode=slice_mode))])
        line_ref.slice_thick_elements(
            slicing_strategies=[xt.Strategy(xt.Teapot(3, mode=slice_mode))])

    line_test['e0'].rot_s_rad = rot_s_rad
    line_test['e0'].shift_x = shift_x
    line_test['e0'].shift_y = shift_y

    p_test = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03)
    p_ref = p_test.copy()

    line_test.build_tracker()
    line_ref.build_tracker()

    line_test.track(p_test)
    line_ref.track(p_ref)

    assert_allclose = np.testing.assert_allclose
    assert_allclose(p_test.x, p_ref.x, rtol=0, atol=1e-13)
    assert_allclose(p_test.px, p_ref.px, rtol=0, atol=1e-13)
    assert_allclose(p_test.y, p_ref.y, rtol=0, atol=1e-13)
    assert_allclose(p_test.py, p_ref.py, rtol=0, atol=1e-13)
    assert_allclose(p_test.zeta, p_ref.zeta, rtol=0, atol=1e-12)
    assert_allclose(p_test.delta, p_ref.delta, rtol=0, atol=1e-13)



import time

import numpy as np

import xtrack
import xdeps as xd

# xt._print.suppress = True

# Load the line
collider = xtrack.Multiline.from_json(
    '../../test_data/hllhc15_collider/collider_00_from_mad.json')
collider.build_trackers()

collider.lhcb1.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['method'] = '4d'
collider.lhcb2.twiss_default['reverse'] = True

class ActionArcPhaseAdvanceFromCell(xd.Action):
    def __init__(self, arc_name, line_name, line):
        assert arc_name in ['12', '23', '34', '45', '56', '67', '78', '81']
        assert line_name in ['lhcb1', 'lhcb2']
        self.arc_name = arc_name
        self.line_name = line_name
        self.line = line

        beam_number = line_name[-1:]
        sector_start_number = arc_name[:1]
        sector_end_number = arc_name[1:]
        self.start_cell = f's.cell.{arc_name}.b{beam_number}'
        self.end_cell = f'e.cell.{arc_name}.b{beam_number}'
        self.start_arc = f'e.ds.r{sector_start_number}.b{beam_number}'
        self.end_arc = f's.ds.l{sector_end_number}.b{beam_number}'

    def run(self):

        twinit_cell = self.line.twiss(
                    start=self.start_cell, end=self.end_cell,
                    init='periodic', only_twiss_init=True)
        #  twinit_cell.element_name is start_cell for b1 and end_cell for b2

        tw_to_end_arc = self.line.twiss(init=twinit_cell,
            start=twinit_cell.element_name, end=self.end_arc)
        tw_to_start_arc = self.line.twiss(init=twinit_cell,
            start=self.start_arc,end=twinit_cell.element_name)

        mux_arc_from_cell = (tw_to_end_arc['mux', self.end_arc]
                             - tw_to_start_arc['mux', self.start_arc])
        muy_arc_from_cell = (tw_to_end_arc['muy', self.end_arc]
                             - tw_to_start_arc['muy', self.start_arc])

        return {
            'mux_arc_from_cell': mux_arc_from_cell,
            'muy_arc_from_cell': muy_arc_from_cell,
            'twinit_cell': twinit_cell,
            'tw_to_end_arc': tw_to_end_arc,
            'tw_to_start_arc': tw_to_start_arc}

class ActionMatchPhaseWithMQTs(xd.Action):

    def __init__(self, arc_name, line_name, line,
                 mux_arc_target, muy_arc_target, restore=False):

        self.action_arc_phase = ActionArcPhaseAdvanceFromCell(
            arc_name=arc_name, line_name=line_name, line=line)
        self.line = line
        self.mux_arc_target = mux_arc_target
        self.muy_arc_target = muy_arc_target
        self.restore = restore

        beam_number = line_name[-1:]
        self.mqt_knob_names = [
            f'kqtf.a{arc_name}b{beam_number}',
            f'kqtd.a{arc_name}b{beam_number}']

    def run(self):
        #store initial knob values
        mqt_knob_values = {
            kk: self.line.vars[kk]._value for kk in self.mqt_knob_names}

        opt = xd.Optimize(
            targets=[
                xd.Target(action=self.action_arc_phase, tar='mux_arc_from_cell',
                            value=self.mux_arc_target, tol=1e-8),
                xd.Target(action=self.action_arc_phase, tar='muy_arc_from_cell',
                            value=self.muy_arc_target, tol=1e-8),
            ],
            vary=[
                xd.VaryList(self.mqt_knob_names, self.line.vars, step=1e-5),
            ])
        opt.solve()

        res = {kk: np.abs(self.line.vars[kk]._value) for kk in self.mqt_knob_names}

        # restore initial knob values
        if self.restore:
            for kk in self.mqt_knob_names:
                self.line.vars[kk] = mqt_knob_values[kk]
        return res


action_arc_phase_s67_b1 = ActionArcPhaseAdvanceFromCell(
                    arc_name='67', line_name='lhcb1', line=collider.lhcb1)
resb1 = action_arc_phase_s67_b1.run()

action_arc_phase_s67_b2 = ActionArcPhaseAdvanceFromCell(
                    arc_name='67', line_name='lhcb2', line=collider.lhcb2)
resb2 = action_arc_phase_s67_b2.run()

# Check for b1
twb1 = collider.lhcb1.twiss()
mux_arc_target_b1 = twb1['mux', 's.ds.l7.b1'] - twb1['mux', 'e.ds.r6.b1']
muy_arc_target_b1 = twb1['muy', 's.ds.l7.b1'] - twb1['muy', 'e.ds.r6.b1']
assert np.isclose(resb1['mux_arc_from_cell'] , mux_arc_target_b1, rtol=1e-6)
assert np.isclose(resb1['muy_arc_from_cell'] , muy_arc_target_b1, rtol=1e-6)

# Check for b2
twb2 = collider.lhcb2.twiss()
mux_arc_target_b2 = twb2['mux', 's.ds.l7.b2'] - twb2['mux', 'e.ds.r6.b2']
muy_arc_target_b2 = twb2['muy', 's.ds.l7.b2'] - twb2['muy', 'e.ds.r6.b2']
assert np.isclose(resb2['mux_arc_from_cell'] , mux_arc_target_b2, rtol=1e-6)
assert np.isclose(resb2['muy_arc_from_cell'] , muy_arc_target_b2, rtol=1e-6)

starting_values = {
    'kqtf.a67b1': collider.vars['kqtf.a67b1']._value,
    'kqtf.a67b2': collider.vars['kqtf.a67b2']._value,
    'kqtd.a67b1': collider.vars['kqtd.a67b1']._value,
    'kqtd.a67b2': collider.vars['kqtd.a67b2']._value,
    'kqf.a67': collider.vars['kqf.a67']._value,
    'kqd.a67': collider.vars['kqd.a67']._value,
}

# # Perturb the quadrupoles
collider.vars['kqtf.a67b1'] = starting_values['kqtf.a67b1'] * 1.1
collider.vars['kqtf.a67b2'] = starting_values['kqtf.a67b2'] * 0.9
collider.vars['kqtd.a67b1'] = starting_values['kqtd.a67b1'] * 0.15
collider.vars['kqtd.a67b2'] = starting_values['kqtd.a67b2'] * 1.15

collider.vars['kqd.a67'] = -0.00872
collider.vars['kqf.a67'] = 0.00877

action_match_mqt_s67_b1 = ActionMatchPhaseWithMQTs(
    arc_name='67', line_name='lhcb1', line=collider.lhcb1,
    mux_arc_target=mux_arc_target_b1, muy_arc_target=muy_arc_target_b1)
action_match_mqt_s67_b2 = ActionMatchPhaseWithMQTs(
    arc_name='67', line_name='lhcb2', line=collider.lhcb2,
    mux_arc_target=mux_arc_target_b2, muy_arc_target=muy_arc_target_b2)


t1 = time.perf_counter()
optimize_phase_arc_67 = xd.Optimize(
    verbose=False,
    assert_within_tol=False,
    solver_options={'n_bisections': 3, 'min_step': 1e-5, 'n_steps_max': 5,},
    targets=[
        xd.Target(action=action_match_mqt_s67_b1, tar='kqtf.a67b1', value=0),
        xd.Target(action=action_match_mqt_s67_b1, tar='kqtd.a67b1', value=0),
        xd.Target(action=action_match_mqt_s67_b2, tar='kqtf.a67b2', value=0),
        xd.Target(action=action_match_mqt_s67_b2, tar='kqtd.a67b2', value=0),
    ],
    vary=[
        xd.Vary(name='kqf.a67', container=collider.vars, step=1e-5),
        xd.Vary(name='kqd.a67', container=collider.vars, step=1e-5),
    ])
optimize_phase_arc_67.solve()

t2 = time.perf_counter()
print(f'Elapsed time: {t2-t1} s')

# Checks
resb1_after = action_arc_phase_s67_b1.run()
tw_init_arcb1 = resb1_after['tw_to_start_arc'].get_twiss_init('e.ds.r6.b1')
twb1_after = collider.lhcb1.twiss(start='e.ds.r6.b1',
                                  end='s.ds.l7.b1',
                                  init=tw_init_arcb1)
assert np.isclose(twb1_after['mux', 's.ds.l7.b1'] - twb1_after['mux', 'e.ds.r6.b1'],
                    mux_arc_target_b1, rtol=0, atol=1e-8)
assert np.isclose(twb1_after['muy', 's.ds.l7.b1'] - twb1_after['muy', 'e.ds.r6.b1'],
                    muy_arc_target_b1, rtol=0, atol=1e-8)
assert np.isclose(twb1_after['betx', 's.cell.67.b1'], twb1_after['betx', 'e.cell.67.b1'],
                    rtol=0, atol=1e-8)
assert np.isclose(twb1_after['bety', 's.cell.67.b1'], twb1_after['bety', 'e.cell.67.b1'],
                    rtol=0, atol=1e-8)

resb2_after = action_arc_phase_s67_b2.run()
tw_init_arcb2 = resb2_after['tw_to_start_arc'].get_twiss_init('e.ds.r6.b2')
twb2_after = collider.lhcb2.twiss(start='e.ds.r6.b2',
                                  end='s.ds.l7.b2',
                                  init=tw_init_arcb2)
assert np.isclose(twb2_after['mux', 's.ds.l7.b2'] - twb2_after['mux', 'e.ds.r6.b2'],
                    mux_arc_target_b2, rtol=0, atol=1e-8)
assert np.isclose(twb2_after['muy', 's.ds.l7.b2'] - twb2_after['muy', 'e.ds.r6.b2'],
                    muy_arc_target_b2, rtol=0, atol=1e-8)
assert np.isclose(twb2_after['betx', 's.cell.67.b2'], twb2_after['betx', 'e.cell.67.b2'],
                    rtol=0, atol=1e-8)
assert np.isclose(twb2_after['bety', 's.cell.67.b2'], twb2_after['bety', 'e.cell.67.b2'],
                    rtol=0, atol=1e-8)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xtrack as xt

fname_line_particles = '../../test_data/hllhc15_noerrors_nobb/line_and_particle.json'
line = xt.Line.from_json(fname_line_particles)
line.particle_ref = xt.Particles(p0c=7e12, mass0=xt.PROTON_MASS_EV)
line.build_tracker()

tw = line.twiss()

R_IP3_IP6 = tw.get_R_matrix(start=0, end='ip6')
R_IP6_IP3 = tw.get_R_matrix(start='ip6', end=len(tw.name)-1)

# # Checks
R_prod = R_IP6_IP3 @ R_IP3_IP6

from xtrack.linear_normal_form import compute_linear_normal_form
eig = np.linalg.eig
norm = np.linalg.norm

R_matrix = tw.R_matrix

W_ref, invW_ref, Rot_ref, _ = compute_linear_normal_form(R_matrix)
W_prod, invW_prod, Rot_prod, _ = compute_linear_normal_form(R_prod)


for i_mode in range(3):
    lam_ref = eig(Rot_ref[2*i_mode:2*i_mode+2, 2*i_mode:2*i_mode+2])[0][0]
    lam_prod = eig(Rot_prod[2*i_mode:2*i_mode+2, 2*i_mode:2*i_mode+2])[0][0]

    assert np.isclose(np.abs(np.angle(lam_ref)) / 2 / np.pi,
                      np.abs(np.angle(lam_prod)) / 2 / np.pi,
                      rtol=0, atol=1e-6)

    assert np.isclose(
        norm(W_prod[:, 2*i_mode] - W_ref[:, 2*i_mode], ord=2)/norm(W_ref[:, 2*i_mode], ord=2),
        0, rtol=0, atol=5e-4)
    assert np.isclose(
        norm(W_prod[:4, 2*i_mode] - W_ref[:4, 2*i_mode], ord=2)/norm(W_ref[:4, 2*i_mode], ord=2),
        0, rtol=0, atol=5e-4)

# Check method=4d

tw4d = line.twiss(method='4d', freeze_longitudinal=True)

R_IP3_IP6_4d = tw4d.get_R_matrix(start=0, end='ip6')
R_IP6_IP3_4d = tw4d.get_R_matrix(start='ip6', end=len(tw4d.name)-1)

R_prod_4d = R_IP6_IP3_4d @ R_IP3_IP6_4d

# Checks
from xtrack.linear_normal_form import compute_linear_normal_form
eig = np.linalg.eig
norm = np.linalg.norm

R_matrix_4d = tw4d.R_matrix

W_ref_4d, invW_ref_4d, Rot_ref_4d, _ = compute_linear_normal_form(
    R_matrix_4d, only_4d_block=True)
W_prod_4d, invW_prod_4d, Rot_prod_4d, _ = compute_linear_normal_form(
    R_prod_4d, only_4d_block=True)

for i_mode in range(3):
    lam_ref_4d = eig(Rot_ref_4d[2*i_mode:2*i_mode+2, 2*i_mode:2*i_mode+2])[0][0]
    lam_prod_4d = eig(Rot_prod_4d[2*i_mode:2*i_mode+2, 2*i_mode:2*i_mode+2])[0][0]

    assert np.isclose(np.abs(np.angle(lam_ref_4d)) / 2 / np.pi,
                      np.abs(np.angle(lam_prod_4d)) / 2 / np.pi,
                      rtol=0, atol=1e-6)

    assert np.isclose(
        norm(W_prod_4d[:, 2*i_mode] - W_ref_4d[:, 2*i_mode], ord=2)/norm(W_ref_4d[:, 2*i_mode], ord=2),
        0, rtol=0, atol=5e-5)





import io
import operator
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple, TypedDict, Union, Literal

from lark import Lark, Transformer, v_args, Token

from xdeps.refs import BaseRef, is_ref

grammar = Path(__file__).with_name('madx.lark').read_text()


# The following types are used to define the output of the parser, to make it
# easier to understand the structure of the parsed MAD-X file and to help
# in spotting mistakes.
VarValueType = Union[int, float, str, bool]


VarType = Union[VarValueType, BaseRef]


class ModifiersType(TypedDict, total=False):
    _repeat: int
    _invert: bool


class ElementType(TypedDict, ModifiersType):
    parent: str
    __extra_items__: Dict[str, VarType]  # Not really supported until PEP 728


class LineType(TypedDict, ModifiersType):
    parent: str  # 'sequence' or 'line'
    l: VarType  # optional, but typing.NotRequired is not available in 3.8
    refer: Literal['centre', 'entry']  # ditto
    elements: List[Tuple[str, Union[ElementType, 'LineType']]]


ElementType = Union[TypedDict('ElementType', {'parent': str}), Dict[str, VarType]]

class MadxOutputType(TypedDict):
    elements: Dict[str, ElementType]
    lines: Dict[str, LineType]
    parameters: Dict[str, ElementType]


@dataclass
class Modifiers:
    repeat: int = 1
    invert: bool = False

    def to_dict(self):
        out = {}
        if self.repeat != 1:
            out['_repeat'] = self.repeat
        if self.invert:
            out['_invert'] = True
        return out


def make_op_handler(op_func):
    return staticmethod(op_func)


def warn(warning):
    print(f'Warning: {warning}')


@v_args(inline=True)
class MadxTransformer(Transformer):
    def __init__(self, vars, functions):
        super().__init__()
        self.elements: Dict[str, ElementType] = {}
        self.lines: Dict[str, LineType] = {}
        self.parameters = {}
        self.vars = vars
        self.functions = functions

    def ignored(self, tokens):
        if tokens:
            statement = ' '.join(str(token) for token in tokens.children)
        else:
            statement = ''
        if statement.startswith('return'):
            return
        if statement == '':
            return
        warn(f'Ignoring statement: `{statement}`')

    def assign_defer(self, name, value) -> Tuple[str, VarType]:
        return name.value.lower(), value

    def assign_value(self, name, value) -> Tuple[str, VarType]:
        if is_ref(value):
            value = value._get_value()
        return name.value.lower(), value

    def special_token_name(self, token):
        return token.value.lower()

    def set_var(self, assignment: Tuple[str, VarType]):
        name, value = assignment
        self.vars[name] = value

    def name_atom(self, name):
        field = name.value.lower()
        if self.vars.default_to_zero and field not in self.vars:
            self.vars[field] = 0
        return self.vars[field]

    def number(self, value):
        float_value = float(value)
        return float_value

    def string_literal(self, string):
        return string.value[1:-1].lower()

    def call(self, function, *args):
        return function(*args)

    def function(self, name_token):
        field = name_token.value.lower()
        return self.functions[field]

    def command(self, name_token, arglist):
        command = name_token.value
        arglist = arglist if isinstance(arglist, list) else [arglist]
        return command.lower(), arglist

    @v_args(inline=False)
    def command_arglist(self, args):
        return args

    def set_flag(self, name_token):
        return name_token.value.lower(), True

    def reset_flag(self, name_token):
        return name_token.value.lower(), False

    def sequence(self, name_token, arglist, *clones) -> Tuple[str, LineType]:
        return name_token.value.lower(), {
            'parent': 'sequence',
            **dict(arglist),
            'elements': list(clones),
        }

    def seqedit(self, arglist, *commands) -> Tuple[str, LineType]:
        (param_name, sequence_name), = arglist

        if param_name != 'sequence':
            raise ValueError(f'Unexpected parameter `{param_name}` of `seqedit`')

        if sequence_name['deferred']:
            raise ValueError('Param `sequence` of `seqedit` cannot be deferred.')

        for command in commands:
            name, params = command

            if any(v['deferred'] for v in params.values()):
                raise ValueError(f'Commands in `seqedit` are not supported with deferred params')

            if name == 'install':
                element_name = params.pop('element')
                self.lines[sequence_name]['elements'].append((element_name, params))
            else:
                warn(f'Command {name} with params {params} is ignored')

    def top_level_sequence(self, sequence):
        name, body = sequence
        self.lines[name] = body

    def clone(self, name_token, command_token, arglist) -> Tuple[str, ElementType]:
        args = dict(arglist)
        parent = command_token.value.lower()

        return name_token.value.lower(), {
            'parent': parent,
            **args,
        }

    def top_level_clone(self, clone):
        name, body = clone
        self.elements[name] = body

    def command_stmt(self, command_token, arglist):
        return command_token.value.lower(), dict(arglist)

    def top_level_command(self, command):
        name, arglist = command
        if name not in self.parameters:
            self.parameters[name] = {}
        self.parameters[name].update(arglist)

    def modifiers(self, *args) -> Modifiers:
        arg_values = set(token.value for token in args)
        modifiers = Modifiers()

        if '-' in arg_values:
            modifiers.invert = True
            arg_values.remove('-')

        if arg_values:
            repeat, = arg_values
            modifiers.repeat = int(repeat)

        return modifiers

    def line_element(self, modifiers, line_item) -> Tuple[str, ElementType]:
        name = None
        body = modifiers.to_dict()
        if isinstance(line_item, Token):
            name = line_item.value.lower()
        elif isinstance(line_item, dict):
            body.update(line_item)
        else:
            raise ValueError(f'Unexpected line element: {line_item}')
        return name, body

    def anonymous_line(self, elements):
        return {
            'parent': 'line',
            'elements': elements,
        }

    def line(self, name_token, anonymous_line) -> Tuple[str, LineType]:
        return name_token.value.lower(), anonymous_line

    def top_level_line(self, line):
        name, body = line
        self.lines[name] = body

    def build_list(self, *args):
        return list(args)

    @v_args(inline=False)
    def start(self, _) -> MadxOutputType:
        return {
            'elements': self.elements,
            'lines': self.lines,
            'parameters': self.parameters,
        }

    def op_arrow(self, a, b):
        a, b = a.lower(), b.lower()
        if b == 'l':
            b = 'length'
        return f'{a}->{b}'

    op_lt = staticmethod(operator.lt)
    op_gt = staticmethod(operator.gt)
    op_le = staticmethod(operator.le)
    op_ge = staticmethod(operator.ge)
    op_eq = staticmethod(operator.eq)
    op_ne = staticmethod(operator.ne)
    op_add = staticmethod(operator.add)
    op_sub = staticmethod(operator.sub)
    op_mul = staticmethod(operator.mul)
    op_div = staticmethod(operator.truediv)
    op_pow = staticmethod(operator.pow)
    op_neg = staticmethod(operator.neg)
    op_pos = staticmethod(operator.pos)


class MadxParser:
    def __init__(self, vars, functions):
        self.transformer = MadxTransformer(vars, functions)
        self.parser = Lark(grammar, parser='lalr', transformer=self.transformer)

    def parse_string(self, text: str) -> MadxOutputType:
        return self.parser.parse(text)  # noqa

    def parse_file(self, path_or_file: Union[str, Path, io.TextIOBase]) -> MadxOutputType:
        if not isinstance(path_or_file, io.TextIOBase):
            file = open(path_or_file)
        else:
            file = path_or_file

        with file:
            return self.parse_string(file.read())


# Run as a script to be able to inspect a MAD-X file as YAML
if __name__ == '__main__':
    import click
    import time
    import yaml

    @click.command()
    @click.argument('file_name')
    @click.option(
        '--output', '-o',
        help='Path to the output file (default: same as input but ending with `.yaml`)',
        default=None,
    )
    @click.option('--test', '-t', is_flag=True, help='Try to load the sequence into Xsuite', default=False)
    def main(file_name, output, test):
        t0 = time.time()
        out = MadxParser().parse_file(file_name)
        print(f"Parsed `{file_name}` in {time.time() - t0} s")

        # This output is for visualisation purposes only: dict ordering is not guaranteed
        # out of the box by the YAML standard (should use !!omap, but it's not supported by PyYAML)
        yaml_out = yaml.dump(out, Dumper=yaml.SafeDumper, sort_keys=False)

        if not output:
            outfile = Path(file_name).with_suffix(suffix='.yaml')
        else:
            outfile = Path(output)

        with outfile.open('w') as f:
            f.write(yaml_out)

        if test:
            from xtrack.mad_parser.loader import MadxLoader
            loader = MadxLoader()
            loader.load_file(outfile)

    main()



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import sixtracktools

import xobjects as xo
import xtrack as xt
import xpart as xp
import xfields as xf


##############
# Build line #
##############

# Read sixtrack input
sixinput = sixtracktools.SixInput(".")
p0c_eV = sixinput.initialconditions[-3] * 1e6

# Build line from sixtrack input
line = sixinput.generate_xtrack_line()

# Info on sixtrack->pyblep conversion
iconv = line.other_info["iconv"]

# Build tracker
tracker = xt.Tracker(line=line)

########################################################
#                  Search closed orbit                 #
# (for comparison purposes we the orbit from sixtrack) #
########################################################

# Load sixtrack tracking data
sixdump_all = sixtracktools.SixDump101("res/dump3.dat")
# Assume first particle to be on the closed orbit
Nele_st = len(iconv)
sixdump_CO = sixdump_all[::2][:Nele_st]
# Get closed-orbit from sixtrack 
part_on_CO = xp.Particles(
        p0c=p0c_eV,
        x=sixdump_CO.x[0],
        px=sixdump_CO.px[0],
        y=sixdump_CO.y[0],
        py=sixdump_CO.py[0],
        zeta=sixdump_CO.tau[0]*sixdump_CO.beta0[0],
        delta=sixdump_CO.delta[0])

print("Closed orbit at start machine:")
for nn in "x px y py zeta delta".split():
    print(f"    {nn}={getattr(part_on_CO, nn)[0]:.5e}:")


#######################################################
#  Store closed orbit and dipole kicks at BB elements #
#######################################################

xf.configure_orbit_dependent_parameters_for_bb(tracker,
                       particle_on_co=part_on_CO)


################
# Save to json #
################
sixdump = sixdump_all[1::2]  # Particle with deviation from CO
part_dict = xp.Particles.from_dict(
        sixdump[0].get_minimal_beam()).to_dict()
part_dict['state'] = 1

with open('line_and_particle.json', 'w') as fid:
    json.dump({
        'line': line.to_dict(),
        'particle': part_dict},
        fid, cls=xo.JEncoder,
        indent=4)




import numpy as np
from xtrack import Line, Node, Multipole

# Or from a sequence definition:
elements = {
    'quad': Multipole(length=0.3, knl=[0, +0.50]),
    'bend': Multipole(length=0.5, knl=[np.pi / 12], hxl=[np.pi / 12]),
}
sequences = {
    'arc': [Node(1.0, 'quad'), Node(4.0, 'bend', from_='quad')],
}
line = Line.from_sequence([
        Node( 0.0, 'arc'),
        Node(10.0, 'arc', name='section2'),
        Node( 3.0, Multipole(knl=[0, 0, 0.1]), from_='section2', name='sext'),
        Node( 3.0, 'quad', name='quad_5', from_='sext'),
    ], length=20,
    elements=elements, sequences=sequences,
    auto_reorder=True, copy_elements=False,
)

line.get_table().show()
# prints:
# #
# name          s element_type isthick isreplica parent_name iscollective
# arc           0 Marker         False     False        None        False
# drift         0 Drift           True     False        None        False
# arcquad       1 Multipole      False     False        None        False
# drift1        1 Drift           True     False        None        False
# arcbend       5 Multipole      False     False        None        False
# drift2        5 Drift           True     False        None        False
# section2     10 Marker         False     False        None        False
# drift3       10 Drift           True     False        None        False
# section2quad 11 Multipole      False     False        None        False
# drift4       11 Drift           True     False        None        False
# sext         13 Multipole      False     False        None        False
# drift5       13 Drift           True     False        None        False
# section2bend 15 Multipole      False     False        None        False
# drift6       15 Drift           True     False        None        False
# quad_5       16 Multipole      False     False        None        False
# drift7       16 Drift           True     False        None        False
# _end_point   20                False     False        None        False


import numpy as np
import xtrack as xt
import xobjects as xo

rdts = ["f4000", "f3100", "f2020", "f1120", 'f1001']

line = xt.Line.from_json(
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')

line['test_dk1'] = 0
line['mb.b32l8.b1'].knl[1] = 'test_dk1'

# sequence_name='dummy'
# mng = line.to_madng(sequence_name=sequence_name)
# mng._sequence_name = sequence_name

# line.build_tracker()
# mvars = MadngVars(mng)
# line.tracker.vars_to_update = [mvars]

# line['a'] = 3.
# assert mng.MADX.a == 3.

tw = line.madng_twiss(normal_form=False)

opt = line.match(
    solve=False,
    vary=[
        xt.Vary('on_x1', step=1e-3),
    ],
    targets=(
        tw.target('px_ng', 50e-6, at='ip1'),
        tw.target('py_ng', 0, at='ip1'),
    ),
)
opt.step(3)
tw_after = line.madng_twiss(normal_form=False)
xo.assert_allclose(tw_after['px_ng', 'ip1'], 50e-6, rtol=5e-3, atol=0)

# tw = line.madng_twiss()
# opt = line.match(
#     solve=False,
#     vary=[
#         xt.VaryList([
#             'kof.a78b1', 'kof.a81b1', 'kof.a12b1', 'kof.a23b1',
#             'kof.a34b1', 'kof.a45b1', 'kof.a56b1', 'kof.a67b1'], step=1e-2),
#         xt.VaryList(
#             ['kod.a78b1', 'kod.a81b1', 'kod.a12b1', 'kod.a23b1',
#              'kod.a34b1', 'kod.a45b1', 'kod.a56b1', 'kod.a67b1'], step=1e-2),
#     ],
#     targets=(
#         tw.target('dqxdjx_nf_ng', 1e6),
#         tw.target('dqydjy_nf_ng', 1e6),
#     ),
# )
# opt.step()


line['on_x1'] = 1.

xo.assert_allclose(line.madng_twiss()['px_ng', 'ip1'], 1e-6, rtol=5e-3, atol=0)

line['on_x1'] = -2.
xo.assert_allclose(line.madng_twiss()['px_ng', 'ip1'], -2e-6, rtol=5e-3, atol=0)

tw_rdt = line.madng_twiss(rdts=rdts)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)

plt.plot(tw_rdt.s, np.abs(tw_rdt.f4000), label='f4000')
plt.plot(tw_rdt.s, np.abs(tw_rdt.f2020), label='f2020')
plt.plot(tw_rdt.s, np.abs(tw_rdt.f1120), label='f1120')
plt.plot(tw_rdt.s, np.abs(tw_rdt.f3100), label='f3100')
plt.xlabel('s [m]')
plt.ylabel(r'|f_{jklm}|')
plt.legend()

plt.show()


import xtrack as xt
import numpy as np

drift = xt.Drift(length=1)

line = xt.Line(elements=[drift])

line.slice_thick_elements(
    slicing_strategies=[xt.Strategy(xt.Uniform(5), element_type=xt.Drift)])
line.build_tracker()
assert line['drift_e0..0'].parent_name == 'e0'
assert line['drift_e0..0']._parent is line['e0']

p0 = xt.Particles(p0c=10e9, x=0.1, px=0.2, y=0.3, py=0.4, delta=0.03)
p_ref = p0.copy()
p_slice = p0.copy()

line.track(p_slice)
drift.track(p_ref)

assert_allclose = np.testing.assert_allclose
assert_allclose = np.testing.assert_allclose
assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

line.to_json('ttt.json')
line2 = xt.Line.from_json('ttt.json')
assert line2['drift_e0..0'].parent_name == 'e0'
assert line2['drift_e0..0']._parent is None

line2.build_tracker()
assert isinstance(line2['drift_e0..0'], xt.DriftSlice)
assert line2['drift_e0..0'].parent_name == 'e0'
assert line2['drift_e0..0']._parent is line2['e0']

line.track(p_slice, backtrack=True)

assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)

line.optimize_for_tracking()

assert isinstance(line['drift_e0..0'], xt.Drift)

p_slice = p0.copy()
line.track(p_slice)

assert_allclose(p_slice.x, p_ref.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p_ref.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p_ref.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p_ref.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p_ref.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p_ref.delta, rtol=0, atol=1e-10)

line.track(p_slice, backtrack=True)

assert_allclose(p_slice.x, p0.x, rtol=0, atol=1e-10)
assert_allclose(p_slice.px, p0.px, rtol=0, atol=1e-10)
assert_allclose(p_slice.y, p0.y, rtol=0, atol=1e-10)
assert_allclose(p_slice.py, p0.py, rtol=0, atol=1e-10)
assert_allclose(p_slice.zeta, p0.zeta, rtol=0, atol=1e-10)
assert_allclose(p_slice.delta, p0.delta, rtol=0, atol=1e-10)


from collections import Counter, UserDict
from collections.abc import Iterable
from functools import cmp_to_key
from typing import Literal
from weakref import WeakSet
from copy import deepcopy
import re
import importlib.util
import sys
import uuid
from pathlib import Path

import numpy as np
import pandas as pd

import xobjects as xo
import xdeps as xd
import xtrack as xt
from xdeps.refs import is_ref
from .multiline_legacy.multiline_legacy import MultilineLegacy
from .progress_indicator import progress

ReferType = Literal['start', 'center', 'centre', 'end']

DEFAULT_REF_STRENGTH_NAME = {
    'Bend': 'k0',
    'Quadrupole': 'k1',
    'Sextupole': 'k2',
    'Octupole': 'k3',
}

def _flatten_components(components, refer: ReferType = 'center'):
    if refer not in ['start', 'center', 'centre', 'end']:
        raise ValueError(
            f'Allowed values for refer are "start", "center" and "end". Got "{refer}".'
        )

    flatt_components = []
    for nn in components:
        if isinstance(nn, Place) and isinstance(nn.name, xt.Line):

            anchor = nn.anchor
            if anchor is None:
                anchor = refer or 'center'

            line = nn.name
            if not line.element_names:
                continue
            sub_components = list(line.element_names).copy()
            if nn.at is not None:
                if isinstance(nn.at, str):
                    at = line._xdeps_eval.eval(nn.at)
                else:
                    at = nn.at
                if anchor=='center' or anchor=='centre':
                    at_of_start_first_element = at - line.get_length() / 2
                elif anchor=='end':
                    at_of_start_first_element = at - line.get_length()
                elif anchor=='start':
                    at_of_start_first_element = at
                else:
                    raise ValueError(f'Unknown anchor {anchor}')
                sub_components[0] = Place(sub_components[0], at=at_of_start_first_element,
                        anchor='start', from_=nn.from_, from_anchor=nn.from_anchor)
            flatt_components += sub_components
        elif isinstance(nn, xt.Line):
            flatt_components += nn.element_names
        elif isinstance(nn, Iterable) and not isinstance(nn, str):
            flatt_components += _flatten_components(nn, refer=refer)
        else:
            flatt_components.append(nn)

    return flatt_components

class Environment:
    def __init__(self, element_dict=None, particle_ref=None, _var_management=None,
                 lines=None):

        '''
        Create an environment.

        Parameters
        ----------
        element_dict : dict, optional
            Dictionary with the elements of the environment.
        particle_ref : ParticleRef, optional
            Reference particle.
        lines : dict, optional
            Dictionary with the lines of the environment.

        Short description of main attributes of the Environment class:
         - Environment[...]: accesses values of variables, elements and lines.
         - ref[...]: provides reference objects to variables and elements.
         - vars[...]: container with all variables, returns variable objects.
         - elements[...]: container with all elements.
         - lines[...]: container with all lines.

        Short description of main methods of the Environment class:
         - info(...): displays information about a variable or element.
         - get(...): returns variable value or element.
         - set(...): sets variable or element properties.
         - eval(...): evaluates an expression provided as a string (returns a value).
         - new_expr(...): creates a new expression from a string.
         - get_expr(...): returns the expression for a variable.
         - new(...): creates a new element.
         - new_line(...): creates a new line.
         - new_builder(...): creates a new builder.
         - place(...): creates a place object, which can be user in new_line(...)
           or by a Builder object.

        Examples
        --------

        .. code-block:: python

            env = xt.Environment()
            env['a'] = 3 # Define a variable
            env.new('mq1', xt.Quadrupole, length=0.3, k1='a')  # Create an element
            env.new('mq2', xt.Quadrupole, length=0.3, k1='-a')  # Create another element

            ln = env.new_line(name='myline', components=[
                'mq',  # Add the element 'mq' at the start of the line
                env.new('mymark', xt.Marker, at=10.0),  # Create a marker at s=10
                env.new('mq1_clone', 'mq1', k1='2*a'),   # Clone 'mq1' with a different k1
                env.place('mq2', at=20.0, from_='mymark'),  # Place 'mq2' at s=20
                ])

        '''
        self._element_dict = element_dict or {}
        self.particle_ref = particle_ref

        if _var_management is not None:
            self._var_management = _var_management
        else:
            self._init_var_management()

        self.lines = EnvLines(self)
        self._lines_weakrefs = WeakSet()
        self._drift_counter = 0
        self.ref = EnvRef(self)

        if lines is not None:

            # Identify common elements
            counts = Counter()
            for ll in lines.values():
                # Extract names of all elements and parents
                elems_and_parents = set(ll.element_names)
                for nn in ll.element_names:
                    if hasattr(ll.element_dict[nn], 'parent_name'):
                        elems_and_parents.add(ll.element_dict[nn].parent_name)
                # Count if it is not a marker or a drift, which will be handled by
                # `import_line`
                for nn in elems_and_parents:
                    if (not (isinstance(ll.element_dict[nn], (xt.Marker))) and
                        not bool(re.match(r'^drift_\d+$', nn))):
                        counts[nn] += 1
            common_elements = [nn for nn, cc in counts.items() if cc>1]

            for nn, ll in lines.items():
                rename_elements = {el: el+'/'+nn for el in common_elements}
                self.import_line(line=ll, suffix_for_common_elements='/'+nn,
                    line_name=nn, rename_elements=rename_elements)
                self.lines[nn]._renamed_elements = rename_elements

        self.metadata = {}

    def __getstate__(self):
        out = self.__dict__.copy()
        out.pop('_lines_weakrefs')
        out.pop('_xdeps_eval_obj', None)
        return out

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._lines_weakrefs = WeakSet()

    def new(self, name, parent, mode=None, at=None, from_=None,
            anchor=None, from_anchor=None,
            extra=None,
            mirror=False, force=False, import_from=None, **kwargs):

        '''
        Create a new element or line.

        Parameters
        ----------
        name : str
            Name of the new element or line
        parent : str or class
            Parent class or name of the parent element
        mode : str, optional
             - clone: clone the parent element or line.
               The parent element or line is copied, together with the associated
               expressions.
             - replica: replicate the parent elements or lines are made.
             - import: clone from a different environment. `import_from` must be
               provided.
        at : float or str, optional
            Position of the created object.
        from_: str, optional
            Name of the element from which the position is calculated (its center
            is used as reference).
        mirror : bool, optional
            Can only be used when cloning lines. If True, the order of the elements
            is reversed.
        import_from : Environment, optional. Only to be used when mode is 'import'.

        Returns
        -------
        str or Place
            Name of the created element or line or a Place object if at or from_ is
            provided.
        '''

        if name in self.element_dict and not force:
            raise ValueError(f'Element `{name}` already exists')

        if from_ is not None or at is not None:
            all_kwargs = locals()
            all_kwargs.pop('self')
            all_kwargs.pop('at')
            all_kwargs.pop('from_')
            all_kwargs.pop('anchor')
            all_kwargs.pop('from_anchor')
            all_kwargs.pop('kwargs')
            all_kwargs.update(kwargs)
            return Place(self.new(**all_kwargs), at=at, from_=from_,
                         anchor=anchor, from_anchor=from_anchor)

        _ALLOWED_ELEMENT_TYPES_IN_NEW = xt.line._ALLOWED_ELEMENT_TYPES_IN_NEW
        _ALLOWED_ELEMENT_TYPES_DICT = xt.line._ALLOWED_ELEMENT_TYPES_DICT
        _STR_ALLOWED_ELEMENT_TYPES_IN_NEW = xt.line._STR_ALLOWED_ELEMENT_TYPES_IN_NEW

        if parent in self.lines:
            parent = self.lines[parent]

        if isinstance(parent, xt.Line):
            assert len(kwargs) == 0, 'No kwargs allowed when creating a line'
            if mode == 'replica':
                assert name is not None, 'Name must be provided when replicating a line'
                return parent.replicate(name=name, mirror=mirror)
            else:
                assert mode in [None, 'clone'], f'Unknown mode {mode}'
                assert name is not None, 'Name must be provided when cloning a line'
                return parent.clone(name=name, mirror=mirror)

        assert mirror is False, 'mirror=True only allowed when cloning lines.'

        if parent is xt.Line or (parent=='Line' and (
            'Line' not in self.lines and 'Line' not in self.element_dict)):
            assert mode is None, 'Mode not allowed when cls is Line'
            return self.new_line(name=name, **kwargs)

        if mode == 'replica':
            assert parent in self.element_dict, f'Element {parent} not found, cannot replicate'
            kwargs['parent_name'] = xo.String(parent)
            parent = xt.Replica
        elif mode == 'clone':
            assert parent in self.element_dict, f'Element {parent} not found, cannot clone'
        else:
            assert mode is None, f'Unknown mode {mode}'

        _eval = self._xdeps_eval.eval

        assert isinstance(parent, str) or parent in _ALLOWED_ELEMENT_TYPES_IN_NEW, (
            'Only '
            + _STR_ALLOWED_ELEMENT_TYPES_IN_NEW
            + ' elements are allowed in `new` for now.')

        needs_instantiation = True
        parent_element = None
        prototype = None
        if isinstance(parent, str):
            if parent in self.element_dict:
                # Clone an existing element
                prototype = parent
                self.element_dict[name] = xt.Replica(parent_name=parent)
                xt.Line.replace_replica(self, name)

                parent_element = self.element_dict[name]
                parent = type(parent_element)
                needs_instantiation = False
            elif parent in _ALLOWED_ELEMENT_TYPES_DICT:
                parent = _ALLOWED_ELEMENT_TYPES_DICT[parent]
                needs_instantiation = True
            else:
                raise ValueError(f'Element type {parent} not found')

        if 'rbend' in kwargs:
            raise ValueError('Use the `RBend` element directly, instead of '
                             'specifying the `rbend` flag.')

        if 'rbarc' in kwargs:
            raise ValueError('Use the `RBend` element with the `length` or '
                             '`length_straight` parameter set accordingly, '
                             'instead of specifying the `rbarc` flag.')

        ref_kwargs, value_kwargs = _parse_kwargs(parent, kwargs, _eval)

        if needs_instantiation: # Parent is a class and not another element
            self.element_dict[name] = parent(**value_kwargs)

        _set_kwargs(name=name, ref_kwargs=ref_kwargs, value_kwargs=value_kwargs,
                    element_dict=self.element_dict, element_refs=self.element_refs)

        if extra is not None:
            assert isinstance(extra, dict)
            self.element_dict[name].extra = extra

        self.element_dict[name].prototype = prototype

        return name

    def _init_var_management(self, dct=None):

        self._var_management = xt.line._make_var_management(element_dict=self.element_dict,
                                               dct=dct)
        self._line_vars = xt.line.LineVars(self)


    def new_line(self, components=None, name=None, refer: ReferType = 'center', length=None):

        '''
        Create a new line.

        Parameters
        ----------
        components : list, optional
            List of components to be added to the line. It can include strings,
            place objects, and lines.
        name : str, optional
            Name of the new line.
        refer : str, optional
            Specifies which part of the component the ``at`` position will refer
            to. Allowed values are ``start``, ``center`` (default; also allowed
            is ``centre```), and ``end``.
        length : float | str, optional
            Length of the line to be built by the builder. Can be an expression.
            If not specified, the length will be the minimum length that can
            fit all the components.

        Returns
        -------
        line
            The new line.

        Examples
        --------
        .. code-block:: python

            env = xt.Environment()
            env['a'] = 3 # Define a variable
            env.new('mq1', xt.Quadrupole, length=0.3, k1='a')  # Create an element
            env.new('mq2', xt.Quadrupole, length=0.3, k1='-a')  # Create another element

            ln = env.new_line(name='myline', components=[
                'mq',  # Add the element 'mq' at the start of the line
                env.new('mymark', xt.Marker, at=10.0),  # Create a marker at s=10
                env.new('mq1_clone', 'mq1', k1='2a'),   # Clone 'mq1' with a different k1
                env.place('mq2', at=20.0, from='mymark'),  # Place 'mq2' at s=20
                ])
        '''

        out = xt.Line(env=self)

        if components is None:
            components = []

        if isinstance(length, str):
            length = self.eval(length)

        components = _resolve_lines_in_components(components, self)
        flattened_components = _flatten_components(components, refer=refer)

        if np.array([isinstance(ss, str) for ss in flattened_components]).all():
            # All elements provided by name
            element_names = [str(ss) for ss in flattened_components]
        else:
            seq_all_places = _all_places(flattened_components)
            tab_unsorted = _resolve_s_positions(seq_all_places, self, refer=refer)
            tab_sorted = _sort_places(tab_unsorted)
            element_names = _generate_element_names_with_drifts(self, tab_sorted,
                                                                length=length)

        out.element_names = element_names
        out._name = name
        out.builder = Builder(env=self, components=components, length=length,
                              name=name, refer=refer)

        # Temporary solution to keep consistency in multiline
        if hasattr(self, '_in_multiline') and self._in_multiline is not None:
            out._var_management = None
            out._in_multiline = self._in_multiline
            out._name_in_multiline = self._name_in_multiline

        self._lines_weakrefs.add(out) # Weak references
        if name is not None:
            self.lines[name] = out

        return out

    def place(self, name, obj=None, at=None, from_=None, anchor=None, from_anchor=None):
        '''
        Create a place object.

        Parameters
        ----------
        name : str or Line
            Name of the element or line to be placed.
        at : float or str, optional
            Position of the created object.
        from_: str, optional
            Name of the element from which the position is calculated (its center
            is used as reference).

        Returns
        -------
        Place
            The new place object.
        '''

        if obj is not None:
            assert not isinstance(name, xt.Line)
            assert name not in self.element_dict
            self.element_dict[name] = obj

        # if list of strings, create a line
        if (isinstance(name, Iterable) and not isinstance(name, str)
            and all(isinstance(item, str) for item in name)):
            name = self.new_line(components=name)

        return Place(name, at=at, from_=from_, anchor=anchor, from_anchor=from_anchor)

    def new_builder(self, components=None, name=None, refer: ReferType = 'center',
                    length=None):
        '''
        Create a new builder.

        Parameters
        ----------
        components : list, optional
            List of components to be added to the builder. It can include strings,
            place objects, and lines.
        name : str, optional
            Name of the line that will be built by the builder.
        refer : str, optional
            Specifies which part of the component the ``at`` position will refer
            to. Allowed values are ``start``, ``center`` (default; also allowed
            is ``centre```), and ``end``.
        length : float | str, optional
            Length of the line to be built by the builder. Can be an expression.
            If not specified, the length will be the minimum length that can
            fit all the components.

        Returns
        -------
        Builder
            The new builder.
        '''

        return Builder(env=self, components=components, name=name, refer=refer,
                       length=length)

    def call(self, filename):
        '''
        Call a file with xtrack commands.

        Parameters
        ----------
        filename : str
            Name of the file to be called.
        '''
        import xtrack
        xtrack._passed_env = self
        try:
            load_module_from_path(Path(filename))
        except Exception as ee:
            xtrack._passed_env = None
            raise ee
        xtrack._passed_env = None

    def copy_element_from(self, name, source, new_name=None):
        return xt.Line.copy_element_from(self, name, source, new_name)


    def _import_element(self, line, name, rename_elements, suffix_for_common_elements,
                        already_imported):
        new_name = name
        if name in rename_elements:
            new_name = rename_elements[name]
        elif (bool(re.match(r'^drift_\d+$', name))
            and line.ref[name].length._expr is None):
            new_name = self._get_a_drift_name()
        elif (name in self.element_dict and
                not (isinstance(line[name], xt.Marker) and
                    isinstance(self.element_dict.get(name), xt.Marker))):
            new_name += suffix_for_common_elements

        self.copy_element_from(name, line, new_name=new_name)
        already_imported[name] = new_name
        if hasattr(line.element_dict[name], 'parent_name'):
            parent_name = line.element_dict[name].parent_name
            if parent_name not in already_imported:
                self._import_element(line, parent_name, rename_elements,
                                     suffix_for_common_elements, already_imported)
            self.element_dict[new_name].parent_name = already_imported[parent_name]

        return new_name

    def import_line(
            self,
            line,
            suffix_for_common_elements=None,
            rename_elements={},
            line_name=None,
            overwrite_vars=False,
    ):
        """Import a line into this environment.

        Parameters
        ----------
        suffix_for_common_elements : str, optional
            Suffix to be added to the names of the elements that are common to
            the imported line and the line in this environment. If None,
            '_{source_line_name}' is used.
        rename_elements : dict, optional
            Dictionary with the elements to be renamed. The keys are the names
            of the elements in `line`, and the values are the new names.
        line_name : str, optional
            Name of the new line. If None, the name of the imported line is used.
        overwrite_vars : bool, optional
            If True, the variables in the imported line will overwrite the
            variables with the same name in this environment. Default is False.
        """
        line_name = line_name or line.name
        if suffix_for_common_elements is None:
            suffix_for_common_elements = f'/{line_name}'

        new_var_values = line.ref_manager.containers['vars']._owner
        if not overwrite_vars:
            new_var_values = new_var_values.copy()
            new_var_values.update(self.ref_manager.containers['vars']._owner)
        self.ref_manager.containers['vars']._owner.update(new_var_values)

        self.ref_manager.copy_expr_from(line.ref_manager, 'vars', overwrite=overwrite_vars)
        old_default_to_zero = self.vars.default_to_zero # Not sure why this is needed
        self.vars.default_to_zero = True
        self.ref_manager.run_tasks()
        self.vars.default_to_zero = old_default_to_zero

        components = []
        already_imported = {}
        for name in line.element_names:
            new_name = self._import_element(
                line, name, rename_elements, suffix_for_common_elements,
                already_imported)

            components.append(new_name)

        out = self.new_line(components=components, name=line_name)

        if line.particle_ref is not None:
            out.particle_ref = line.particle_ref.copy()

        out.config.clear()
        out.config.update(line.config.copy())
        out._extra_config.update(line._extra_config.copy())
        out.metadata.clear()
        out.metadata.update(line.metadata)

        if out.energy_program is not None:
            out.energy_program.line = out


    def _ensure_tracker_consistency(self, buffer):
        for ln in self._lines_weakrefs:
            if ln._has_valid_tracker() and ln._buffer is not buffer:
                ln.discard_tracker()

    def _get_a_drift_name(self):
        self._drift_counter += 1
        nn = f'drift_{self._drift_counter}'
        if nn not in self.element_dict:
            return nn
        else:
            return self._get_a_drift_name()

    def __setitem__(self, key, value):

        if isinstance(value, xt.Line):
            assert value.env is self, 'Line must be in the same environment'
            if key in self.lines:
                raise ValueError(f'There is already a line with name {key}')
            if key in self.element_dict:
                raise ValueError(f'There is already an element with name {key}')
            self.lines[key] = value
        else:
            xt.Line.__setitem__(self, key, value)

    def to_dict(self, include_var_management=True):

        out = {}
        out["elements"] = {k: el.to_dict() for k, el in self.element_dict.items()}

        if self.particle_ref is not None:
            out['particle_ref'] = self.particle_ref.to_dict()
        if self._var_management is not None and include_var_management:
            if hasattr(self, '_in_multiline') and self._in_multiline is not None:
                raise ValueError('The line is part ot a MultiLine object. '
                    'To save without expressions please use '
                    '`line.to_dict(include_var_management=False)`.\n'
                    'To save also the deferred expressions please save the '
                    'entire multiline.\n ')

            out.update(self._var_management_to_dict())

        if hasattr(self, '_bb_config') and self._bb_config is not None:
            out['_bb_config'] = {}
            for nn, vv in self._bb_config.items():
                if nn == 'dataframes':
                    out['_bb_config'][nn] = {}
                    for kk, vv in vv.items():
                        if vv is not None:
                            out['_bb_config'][nn][kk] = vv.to_dict()
                        else:
                            out['_bb_config'][nn][kk] = None
                else:
                    out['_bb_config'][nn] = vv

        out["metadata"] = deepcopy(self.metadata)

        out['xsuite_data_type'] = 'Environment'

        out['lines'] = {}

        for nn, ll in self.lines.items():
            out['lines'][nn] = ll.to_dict(include_element_dict=False,
                                        include_var_management=False)

        return out

    @classmethod
    def from_dict(cls, dct):
        cls = xt.Environment

        ldummy = xt.Line.from_dict(dct)
        out = cls(element_dict=ldummy.element_dict, particle_ref=ldummy.particle_ref,
                _var_management=ldummy._var_management)
        out._line_vars = xt.line.LineVars(out)

        for nn in dct['lines'].keys():
            ll = xt.Line.from_dict(dct['lines'][nn], env=out, verbose=False)
            out[nn] = ll

        if '_bb_config' in dct:
            out._bb_config = dct['_bb_config']
            for nn, vv in dct['_bb_config']['dataframes'].items():
                if vv is not None:
                    df = pd.DataFrame(vv)
                else:
                    df = None
                out._bb_config[
                    'dataframes'][nn] = df

        if "metadata" in dct:
            out.metadata = dct["metadata"]

        return out

    @classmethod
    def from_json(cls, file, **kwargs):

        """Constructs an environment from a json file.

        Parameters
        ----------
        file : str or file-like object
            Path to the json file or file-like object.
            If filename ends with '.gz' file is decompressed.
        **kwargs : dict
            Additional keyword arguments passed to `Environment.from_dict`.

        Returns
        -------
        environment : Environment
            Environment object.

        """

        dct = xt.json.load(file)

        return cls.from_dict(dct, **kwargs)


    def to_json(self, file, indent=1, **kwargs):
        '''Save the environment to a json file.

        Parameters
        ----------
        file: str or file-like object
            The file to save to. If a string is provided, a file is opened and
            closed. If a file-like object is provided, it is used directly.
        **kwargs:
            Additional keyword arguments are passed to the `Environment.to_dict` method.

        '''

        xt.json.dump(self.to_dict(**kwargs), file, indent=indent)

    @classmethod
    def from_madx(cls, filename=None, madx=None, stdout=None, return_lines=False, **kwargs):
        '''
        Load a multiline from a MAD-X file.

        Parameters
        ----------
        file: str
            The MAD-X file to load from.
        **kwargs: dict
            Additional keyword arguments are passed to the `Line.from_madx_sequence`
            method.

        Returns
        -------
        new_multiline: Multiline
            The multiline object.
        '''
        return xt.multiline_legacy._multiline_from_madx(cls, filename=filename, madx=madx, stdout=stdout,
                             return_lines=return_lines, **kwargs)

    @property
    def elements(self):
        return self.element_dict

    @property
    def line_names(self):
        return list(self.lines.keys())

    @property
    def functions(self):
        return self._xdeps_fref

    def _remove_element(self, name):

        pars_with_expr = list(
            self._xdeps_manager.tartasks[self.element_refs[name]].keys())

        # Kill all references
        for rr in pars_with_expr:
            if isinstance(rr, xd.refs.AttrRef):
                setattr(self[name], rr._key, 99)
            elif isinstance(rr, xd.refs.ItemRef):
                getattr(self[name], rr._owner._key)[rr._key] = 99
            else:
                raise ValueError('Only AttrRef and ItemRef are supported for now')

        self.element_dict.pop(name)

    def __getattr__(self, key):
        if key == 'lines':
            return object.__getattribute__(self, 'lines')
        if key in self.lines:
            return self.lines[key]
        else:
            raise AttributeError(f"Environment object has no attribute `{key}`.")

    def __dir__(self):
        return [nn for nn  in list(self.lines.keys()) if '.' not in nn
                    ] + object.__dir__(self)

    def set_multipolar_errors(env, errors):

        '''
        Set multipolar errors for specified elements of the environment.

        Parameters
        ----------

        errors : dict
            Dictionary with the errors to be set. The keys are the names of the
            elements, and the values are dictionaries with the following keys:
             - rel_knl: list of relative errors for the normal multipolar strengths.
             - rel_ksl: list of relative errors for the skew multipolar strengths.
             - refer: name of the strength to be used as reference, which is
               multiplied by the length. If None, the default reference strength
               is used (k0 for bends, k1 for quadrupoles, k2 for sextupoles,
               and k3 for octupoles).

        Examples
        --------

        .. code-block:: python

            env = xt.Environment()
            env.vars.default_to_zero = True
            line = env.new_line(components=[
                env.new('mq', 'Quadrupole', length=0.5, k1='kq'),
                env.new('mqs', 'Quadrupole', length=0.5, k1s='kqs'),
                env.new('mb', 'Bend', length=0.5, angle='ang', k0_from_h=True),
            ])

            env.set_multipolar_errors({
                'mq': {'rel_knl': [1e-6, 1e-5, 1e-4],
                       'rel_ksl': [-1e-6, -1e-5, -1e-4]},
                'mqs': {'rel_knl': [2e-6, 2e-5, 2e-4],
                        'rel_ksl': [3e-6, 3e-5, 3e-4],
                        'refer': 'k1s'},
                'mb': {'rel_knl': [2e-6, 3e-5, 4e-4],
                       'rel_ksl': [5e-6, 6e-5, 7e-4]},
                })

        '''

        for ele_name in progress(errors.keys(), desc='Setting multipolar errors'):

            err = errors[ele_name]
            rel_knl = err.get('rel_knl', [])
            rel_ksl = err.get('rel_ksl', [])
            refer = err.get('refer', None)
            ele_class = env[ele_name].__class__.__name__

            if 'Replica' in ele_class or 'Slice' in ele_class:
                raise ValueError(f'Cannot set multipolar errors for element `{ele_name}`'
                                 f' of type `{ele_class}`')

            if refer is not None:
                reference_strength_name = refer
            else:
                reference_strength_name = DEFAULT_REF_STRENGTH_NAME.get(ele_class, None)

            if reference_strength_name is None:
                raise ValueError(f'Cannot find reference strength for element `{ele_name}`')

            ref_str_ref = getattr(env.ref[ele_name], reference_strength_name)
            length_ref = env.ref[ele_name].length

            for ii, kk in enumerate(rel_knl):
                err_vname = f'err_{ele_name}_knl{ii}'
                env[err_vname] = kk
                if (env.ref[ele_name].knl[ii]._expr is None or env.ref[err_vname] in
                        env.ref[ele_name].knl[ii]._expr._get_dependencies()):
                    env[ele_name].knl[ii] += env.ref[err_vname] * ref_str_ref * length_ref

            for ii, kk in enumerate(rel_ksl):
                err_vname = f'err_{ele_name}_ksl{ii}'
                env[err_vname] = kk
                if (env.ref[ele_name].ksl[ii]._expr is None or env.ref[err_vname] in
                        env.ref[ele_name].ksl[ii]._expr._get_dependencies()):
                    env[ele_name].ksl[ii] += env.ref[err_vname] * ref_str_ref * length_ref

    element_dict = xt.Line.element_dict
    _xdeps_vref = xt.Line._xdeps_vref
    _xdeps_fref = xt.Line._xdeps_fref
    _xdeps_manager = xt.Line._xdeps_manager
    _xdeps_eval = xt.Line._xdeps_eval
    element_refs = xt.Line.element_refs
    vars = xt.Line.vars
    varval = xt.Line.varval
    vv = xt.Line.vv
    __getitem__ = xt.Line.__getitem__
    set = xt.Line.set
    get = xt.Line.get
    eval = xt.Line.eval
    info = xt.Line.info
    get_expr = xt.Line.get_expr
    new_expr = xt.Line.new_expr
    ref_manager = xt.Line.ref_manager
    _var_management_to_dict = xt.Line._var_management_to_dict

    twiss = MultilineLegacy.twiss
    discard_trackers = MultilineLegacy.discard_trackers
    build_trackers = MultilineLegacy.build_trackers
    match = MultilineLegacy.match
    match_knob = MultilineLegacy.match_knob
    install_beambeam_interactions = MultilineLegacy.install_beambeam_interactions
    configure_beambeam_interactions =  MultilineLegacy.configure_beambeam_interactions
    apply_filling_pattern = MultilineLegacy.apply_filling_pattern


class Place:

    def __init__(self, name, at=None, from_=None, anchor=None, from_anchor=None):

        if isinstance(at, str) and '@' in at:
            at_parts = at.split('@')
            assert len(at_parts) == 2
            assert from_ is None
            assert from_anchor is None
            at = 0
            from_ = at_parts[0]
            from_anchor = at_parts[1]

        if from_ is not None:
            assert isinstance(from_, str)
            if '@' in from_:
                from_parts = from_.split('@')
                assert len(from_parts) == 2
                from_ = from_parts[0]
                from_anchor = from_parts[1]

        assert anchor in [None, 'center', 'centre', 'start', 'end']
        assert from_anchor in [None, 'center', 'centre', 'start', 'end']

        self.name = name
        self.at = at
        self.from_ = from_
        self.anchor = anchor
        self.from_anchor = from_anchor

    def __repr__(self):
        out = f'Place({self.name}'
        if self.at is not None: out += f', at={self.at}'
        if self.from_ is not None: out += f', from_={self.from_}'
        if self.anchor is not None: out += f', anchor={self.anchor}'
        if self.from_anchor is not None: out += f', from_anchor={self.from_anchor}'

        out += ')'
        return out

    def copy(self):
        out = Place('dummy')
        out.__dict__ = self.__dict__.copy()
        return out

def _all_places(seq):
    seq_all_places = []
    for ss in seq:
        if isinstance(ss, Place):
            seq_all_places.append(ss)
        elif not isinstance(ss, str) and hasattr(ss, '__iter__'):
            # Find first place
            i_first = None
            for ii, sss in enumerate(ss):
                if isinstance(sss, Place):
                    i_first = ii
                    break
                assert isinstance(sss, str) or isinstance(sss, xt.Line), (
                    'Only places, elements, strings or Lines are allowed in sequences')
            ss_aux = _all_places(ss)
            seq_all_places.extend(ss_aux)
        else:
            assert isinstance(ss, str) or isinstance(ss, xt.Line), (
                'Only places, elements, strings or Lines are allowed in sequences')
            seq_all_places.append(Place(ss, at=None, from_=None))
    return seq_all_places

# In case we want to allow for the length to be an expression
# def _length_expr_or_val(name, line):
#     if isinstance(line[name], xt.Replica):
#         name = line[name].resolve(line, get_name=True)

#     if not line[name].isthick:
#         return 0

#     if line.element_refs[name]._expr is not None:
#         return line.element_refs[name]._expr
#     else:
#         return line[name].length

def _compute_one_s(at, anchor, from_anchor, self_length, from_length, s_start_from,
                   default_anchor):

    if is_ref(at):
        at = at._value

    if anchor is None:
        anchor = default_anchor

    if from_anchor is None:
        from_anchor = default_anchor

    s_from = 0
    if from_length is not None:
        s_from = s_start_from
        if from_anchor == 'center' or from_anchor == 'centre':
            s_from += from_length / 2
        elif from_anchor == 'end':
            s_from += from_length

    ds_self = 0
    if anchor == 'center' or anchor=='centre':
        ds_self = self_length / 2
    elif anchor == 'end':
        ds_self = self_length

    s_start_self = s_from + at - ds_self

    return s_start_self

def _resolve_s_positions(seq_all_places, env, refer: ReferType = 'center',
                         allow_duplicate_places=True, s_tol=1e-10):

    if not allow_duplicate_places:
        raise NotImplementedError('allow_duplicate_places=False not yet implemented')

    seq_all_places = [ss.copy() for ss in seq_all_places]
    names_unsorted = [ss.name for ss in seq_all_places]

    aux_line = env.new_line(components=names_unsorted, refer=refer)

    # Prepare table for output
    tt_out = aux_line.get_table()
    tt_out['length'] = np.diff(tt_out.s, append=tt_out.s[-1])
    tt_out = tt_out.rows[:-1] # Remove endpoint

    tt_lengths = xt.Table({'name': tt_out.env_name, 'length': tt_out.length})

    s_start_for_place = {}  # start positions
    place_for_name = {}
    n_resolved = 0
    n_resolved_prev = -1

    assert len(seq_all_places) == len(set(seq_all_places)), 'Duplicate places detected'

    if seq_all_places[0].at is None:
        # In case we want to allow for the length to be an expression
        s_start_for_place[seq_all_places[0]] = 0
        place_for_name[seq_all_places[0].name] = seq_all_places[0]
        n_resolved += 1

    while n_resolved != n_resolved_prev:
        n_resolved_prev = n_resolved
        for ii, ss in enumerate(seq_all_places):

            if ss in s_start_for_place:  # Already resolved
                continue

            if ss.from_ is not None or ss.from_anchor is not None:
                if ss.at is None:
                    raise ValueError(
                        f'Cannot specify `from_ `or `from_anchor` without providing `at`.'
                        f'Error in place `{ss}`.')

            if ss.at is None:
                ss_prev = seq_all_places[ii-1]
                if ss_prev in s_start_for_place:
                    s_start_for_place[ss] = (s_start_for_place[ss_prev]
                                             + tt_lengths['length', ss_prev.name])
                    place_for_name[ss.name] = ss
                    ss.at = 0
                    ss.from_ = ss_prev.name
                    ss.from_anchor = 'end'
                    n_resolved += 1
            else:
                if isinstance(ss.at, str):
                    at = aux_line._xdeps_eval.eval(ss.at)
                else:
                    at = ss.at

                from_length=None
                s_start_from=None
                if ss.from_ is not None:
                    if ss.from_ not in place_for_name:
                        continue # Cannot resolve yet
                    else:
                        from_length = tt_lengths['length', ss.from_]
                        s_start_from=s_start_for_place[place_for_name[ss.from_]]

                s_start_for_place[ss] = _compute_one_s(at, anchor=ss.anchor,
                    from_anchor=ss.from_anchor,
                    self_length=tt_lengths['length', ss.name],
                    from_length=from_length,
                    s_start_from=s_start_from,
                    default_anchor=refer)

                place_for_name[ss.name] = ss
                n_resolved += 1

    if n_resolved != len(seq_all_places):
        unresolved_pos = set(seq_all_places) - set(s_start_for_place.keys())
        raise ValueError(f'Could not resolve all s positions: {unresolved_pos}')

    if n_resolved != len(seq_all_places):
        unresolved_pos = set(seq_all_places) - set(s_start_for_place.keys())
        raise ValueError(f'Could not resolve all s positions: {unresolved_pos}')

    aux_s_start = np.array([s_start_for_place[ss] for ss in seq_all_places])
    aux_s_center = aux_s_start + tt_out['length'] / 2
    aux_s_end = aux_s_start + tt_out['length']
    tt_out['s_start'] = aux_s_start
    tt_out['s_center'] = aux_s_center
    tt_out['s_end'] = aux_s_end

    tt_out['from_'] = np.array([ss.from_ for ss in seq_all_places])
    tt_out['from_anchor'] = np.array([ss.from_anchor for ss in seq_all_places])

    return tt_out

# @profile
def _sort_places(tt_unsorted, s_tol=1e-10, allow_non_existent_from=False):

    tt_unsorted['i_place'] = np.arange(len(tt_unsorted))

    # Sort by s_center
    iii = _argsort_s(tt_unsorted.s_center, tol=s_tol)
    tt_s_sorted = tt_unsorted.rows[iii]

    group_id = np.zeros(len(tt_s_sorted), dtype=int)
    group_id[0] = 0
    for ii in range(1, len(tt_s_sorted)):
        if abs(tt_s_sorted.s_center[ii] - tt_s_sorted.s_center[ii-1]) > s_tol:
            group_id[ii] = group_id[ii-1] + 1
        elif tt_s_sorted.isthick[ii]: # Needed in Line.insert (on the first sorting pass there can be overlapping elements)
            group_id[ii] = group_id[ii-1] + 1
        else:
            group_id[ii] = group_id[ii-1]

    tt_s_sorted['group_id'] = group_id
    # tt_s_sorted.show(cols=['group_id', 's_center', 'name', 'from_', 'from_anchor', 'i_place'])

    # cache indices (indices will change but only within groups, so no need to update in the loop)
    # This trick gives me x40 speedup compared to using tt_s_sorted.rows.indices
    # at each iteration.
    ind_name = {nn: ii for ii, nn in enumerate(tt_s_sorted.name)}

    n_places = len(tt_s_sorted)
    i_start_group = 0
    i_place_sorted = []
    while i_start_group < n_places:
        i_group = tt_s_sorted['group_id', i_start_group]
        i_end_group = i_start_group + 1
        while i_end_group < n_places and tt_s_sorted['group_id', i_end_group] == i_group:
            i_end_group += 1
        # print(f'Group {i_group}: {tt_s_sorted.name[i_start_group:i_end_group]}')

        n_group = i_end_group - i_start_group
        if n_group == 1: # Single element
            i_place_sorted.append(tt_s_sorted.i_place[i_start_group])
            i_start_group = i_end_group
            continue

        if np.all(tt_s_sorted.from_anchor[i_start_group:i_end_group] == None): # Nothing to do
            i_place_sorted.extend(list(tt_s_sorted.i_place[i_start_group:i_end_group]))
            i_start_group = i_end_group

        tt_group = tt_s_sorted.rows[i_start_group:i_end_group]
        # tt_group.show(cols=['s_center', 'name', 'from_', 'from_anchor'])

        for ff in tt_group.from_:
            if ff is None:
                continue
            if ff not in ind_name:
                if allow_non_existent_from:
                    continue
                else:
                    raise ValueError(f'Element {ff} not found in the line')
            i_from_global = ind_name[ff] - i_start_group
            key_sort = np.zeros(n_group, dtype=int)

            if i_from_global < 0:
                key_sort[:] = 2
            elif i_from_global >= n_group:
                key_sort[:] = -2
            else:
                i_local = tt_group.rows.indices[ff][0] # I need to use this because it might change in the group resortings
                key_sort[i_local] = 0
                key_sort[:i_local] = -2
                key_sort[i_local+1:] = 2

            from_present = tt_group['from_']
            from_anchor_present = tt_group['from_anchor']

            mask_pack_before = (from_present == ff) & (from_anchor_present == 'start')
            mask_pack_after = (from_present == ff) & (from_anchor_present == 'end')
            key_sort[mask_pack_before] = -1
            key_sort[mask_pack_after] = 1

            if np.all(np.diff(key_sort) >=0):
                continue # already sorted
            tt_group = tt_group.rows[np.argsort(key_sort, kind='stable')]

        i_place_sorted.extend(list(tt_group.i_place))
        i_start_group = i_end_group

    tt_sorted = tt_unsorted.rows[i_place_sorted]

    tt_sorted['s_center'] = tt_sorted['s_start'] + tt_sorted['length'] / 2
    tt_sorted['s_end'] = tt_sorted['s_start'] + tt_sorted['length']

    tt_sorted['ds_upstream'] = 0 * tt_sorted['s_start']
    tt_sorted['ds_upstream'][1:] = tt_sorted['s_start'][1:] - tt_sorted['s_end'][:-1]
    tt_sorted['ds_upstream'][0] = tt_sorted['s_start'][0]
    tt_sorted['s'] = tt_sorted['s_start']

    return tt_sorted

def _generate_element_names_with_drifts(env, tt_sorted, length=None, s_tol=1e-6):

    names_with_drifts = []
    # Create drifts
    for ii, nn in enumerate(tt_sorted.env_name):
        ds_upstream = tt_sorted['ds_upstream', ii]
        if np.abs(ds_upstream) > s_tol:
            assert ds_upstream > 0, f'Negative drift length: {ds_upstream}, upstream of {nn}'
            drift_name = env._get_a_drift_name()
            env.new(drift_name, xt.Drift, length=ds_upstream)
            names_with_drifts.append(drift_name)
        names_with_drifts.append(nn)

    if length is not None:
        length_line = tt_sorted['s_end'][-1]
        if length_line > length + s_tol:
            raise ValueError(f'Line length {length_line} is greater than the requested length {length}')
        if length_line < length - s_tol:
            drift_name = env._get_a_drift_name()
            env.new(drift_name, xt.Drift, length=length - length_line)
            names_with_drifts.append(drift_name)

    return list(map(str, names_with_drifts))

def _parse_kwargs(cls, kwargs, _eval):
    ref_kwargs = {}
    value_kwargs = {}
    for kk in kwargs:
        if hasattr(kwargs[kk], '_value'):
            ref_kwargs[kk] = kwargs[kk]
            value_kwargs[kk] = kwargs[kk]._value
        elif (hasattr(cls, '_xofields') and kk in cls._xofields
                and xo.array.is_array(cls._xofields[kk])):
            assert hasattr(kwargs[kk], '__iter__'), (
                f'{kk} should be an iterable for {cls} element')
            ref_vv = []
            value_vv = []
            for ii, vvv in enumerate(kwargs[kk]):
                if hasattr(vvv, '_value'):
                    ref_vv.append(vvv)
                    value_vv.append(vvv._value)
                elif isinstance(vvv, str):
                    ref_vv.append(_eval(vvv))
                    if hasattr(ref_vv[-1], '_value'):
                        value_vv.append(ref_vv[-1]._value)
                    else:
                        value_vv.append(ref_vv[-1])
                else:
                    ref_vv.append(None)
                    value_vv.append(vvv)
            ref_kwargs[kk] = ref_vv
            value_kwargs[kk] = value_vv
        elif (isinstance(kwargs[kk], str) and hasattr(cls, '_xofields')
            and (not hasattr(cls, '_noexpr_fields') or kk not in cls._noexpr_fields)):
            ref_kwargs[kk] = _eval(kwargs[kk])
            if hasattr(ref_kwargs[kk], '_value'):
                value_kwargs[kk] = ref_kwargs[kk]._value
            else:
                value_kwargs[kk] = ref_kwargs[kk]
        elif isinstance(kwargs[kk], xo.String):
            vvv = kwargs[kk].to_str()
            value_kwargs[kk] = vvv
        else:
            value_kwargs[kk] = kwargs[kk]

    return ref_kwargs, value_kwargs

def _set_kwargs(name, ref_kwargs, value_kwargs, element_dict, element_refs):
    for kk in value_kwargs:
        if hasattr(value_kwargs[kk], '__iter__') and not isinstance(value_kwargs[kk], str):
            len_value = len(value_kwargs[kk])
            getattr(element_dict[name], kk)[:len_value] = value_kwargs[kk]
            if kk in ref_kwargs:
                for ii, vvv in enumerate(value_kwargs[kk]):
                    if ref_kwargs[kk][ii] is not None:
                        getattr(element_refs[name], kk)[ii] = ref_kwargs[kk][ii]
        elif kk in ref_kwargs:
            setattr(element_refs[name], kk, ref_kwargs[kk])
        else:
            setattr(element_dict[name], kk, value_kwargs[kk])

class EnvRef:
    def __init__(self, env):
        self.env = env

    def __getitem__(self, name):
        if hasattr(self.env, 'lines') and name in self.env.lines:
            return self.env.lines[name].ref
        elif name in self.env.element_dict:
            return self.env.element_refs[name]
        elif name in self.env.vars:
            return self.env.vars[name]
        else:
            raise KeyError(f'Name {name} not found.')

    def __setitem__(self, key, value):
        if isinstance(value, xt.Line):
            assert value.env is self.env, 'Line must be in the same environment'
            if key in self.env.lines:
                raise ValueError(f'There is already a line with name {key}')
            if key in self.env.element_dict:
                raise ValueError(f'There is already an element with name {key}')
            self.env.lines[key] = value

        if hasattr(value, '_value'):
            val_ref = value
            val_value = value._value
        else:
            val_ref = value
            val_value = value

        if np.isscalar(val_value):
            if key in self.env.element_dict:
                raise ValueError(f'There is already an element with name {key}')
            self.env.vars[key] = val_ref
        else:
            if key in self.env.vars:
                raise ValueError(f'There is already a variable with name {key}')
            self.element_refs[key] = val_ref


class Builder:
    def __init__(self, env, components=None, name=None, length=None, refer: ReferType = 'center'):
        self.env = env
        self.components = components or []
        self.name = name
        self.refer = refer
        self.length = length

    def __repr__(self):
        parts = [f'name={self.name!r}']
        if self.length is not None:
            parts.append(f'length={self.length!r}')
        if self.refer not in {'center', 'centre'}:
            parts.append(f'refer={self.refer!r}')
        parts.append(f'components={self.components!r}')
        args_str = ', '.join(parts)
        return f'Builder({args_str})'

    def new(self, name, cls, at=None, from_=None, extra=None, force=False,
            **kwargs):
        out = self.env.new(
            name, cls, at=at, from_=from_, extra=extra, force=force, **kwargs)
        self.components.append(out)
        return out

    def place(self, name, obj=None, at=None, from_=None, anchor=None, from_anchor=None):
        out = self.env.place(name=name, obj=obj, at=at, from_=from_,
                             anchor=anchor, from_anchor=from_anchor)
        self.components.append(out)
        return out

    def build(self, name=None):
        if name is None:
            name = self.name
        out =  self.env.new_line(components=self.components, name=name, refer=self.refer,
                                 length=self.length)
        out.builder = self
        return out

    def set(self, *args, **kwargs):
        self.components.append(self.env.set(*args, **kwargs))

    def get(self, *args, **kwargs):
        return self.env.get(*args, **kwargs)


    def resolve_s_positions(self):
        components = self.components
        if components is None:
            components = []

        components = _resolve_lines_in_components(components, self.env)
        flattened_components = _flatten_components(components, refer=self.refer)

        seq_all_places = _all_places(flattened_components)
        tab_unsorted = _resolve_s_positions(seq_all_places, self.env, refer=self.refer)
        tab_sorted = _sort_places(tab_unsorted)
        return tab_sorted

    def flatten(self, inplace=False):

        assert not inplace, 'Inplace not yet implemented'

        out = self.__class__(self.env)
        out.__dict__.update(self.__dict__)

        components = _resolve_lines_in_components(self.components, self.env)
        out.components = _flatten_components(components, refer=self.refer)
        out.components = _all_places(out.components)
        return out

    @property
    def element_dict(self):
        return self.env.element_dict

    @property
    def ref(self):
        return self.env.ref

    @property
    def vars(self):
        return self.env.vars

    def __getitem__(self, key):
        return self.env[key]

    def __setitem__(self, key, value):
        self.env[key] = value


class EnvLines(UserDict):

    def __init__(self, env):
        self.data = {}
        self.env = env

    def __setitem__(self, key, value):
        self.env._lines_weakrefs.add(value)
        UserDict.__setitem__(self, key, value)

def get_environment(verbose=False):
    import xtrack
    if hasattr(xtrack, '_passed_env') and xtrack._passed_env is not None:
        if verbose:
            print('Using existing environment')
        return xtrack._passed_env
    else:
        if verbose:
            print('Creating new environment')
        return Environment()

def _argsort_s(seq, tol=10e-10):
    """Argsort, but with a tolerance; `sorted` is stable."""
    seq_indices = np.arange(len(seq))

    def comparator(i, j):
        a, b = seq[i], seq[j]
        if np.abs(a - b) < tol:
            return 0
        return -1 if a < b else 1

    return sorted(seq_indices, key=cmp_to_key(comparator))


def load_module_from_path(file_path):
    """
    Load a module from the given file path, always generating a unique module name.

    Parameters:
        file_path (str): The full path to the module file.

    Returns:
        module: The newly loaded module.
    """
    # Generate a unique module name using uuid4.
    module_name = f"module_{uuid.uuid4().hex}"

    # Create a module spec from the file location.
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    if spec is None:
        raise ImportError(f"Cannot load module from path: {file_path}")

    # Create a new module based on the spec.
    module = importlib.util.module_from_spec(spec)

    # Execute the module in its own namespace.
    spec.loader.exec_module(module)

    return module

def _reverse_element(env, name):
    """Return a reversed element without modifying the original."""

    SUPPORTED = {'RBend', 'Bend', 'Quadrupole', 'Sextupole', 'Octupole',
                'Multipole', 'Cavity', 'Solenoid', 'RFMultipole',
                'Marker', 'Drift', 'LimitRect', 'LimitEllipse', 'LimitPolygon',
                'LimitRectEllipse'}

    ee = env.get(name)
    ee_ref = env.ref[name]

    if ee.__class__.__name__ not in SUPPORTED:
        raise NotImplementedError(
            f'Cannot reverse the element `{name}`, as reversing elements '
            f'of type `{ee.__class__.__name__}` is not supported!'
        )

    def _reverse_field(key):
        if hasattr(ee, key):
            key_ref = getattr(ee_ref, key)
            if key_ref._expr is not None:
                setattr(ee_ref, key,  -(key_ref._expr))
            else:
                setattr(ee_ref, key,  -(key_ref._value))

    def _exchange_fields(key1, key2):
        value1 = None
        if hasattr(ee, key1):
            key1_ref = getattr(ee_ref, key1)
            value1 = key1_ref._expr or key1_ref._value

        value2 = None
        if hasattr(ee, key2):
            key2_ref = getattr(ee_ref, key2)
            value2 = key2_ref._expr or key2_ref._value


        if value1 is not None:
            setattr(ee_ref, key2, value1)

        if value2 is not None:
            setattr(ee_ref, key1, value2)

    _reverse_field('k0s')
    _reverse_field('k1')
    _reverse_field('k2s')
    _reverse_field('k3')
    _reverse_field('ks')
    _reverse_field('ksi')
    _reverse_field('rot_s_rad')

    if hasattr(ee, 'lag'):
        ee_ref.lag = 180 - (ee_ref.lag._expr or ee_ref.lag._value)

    if hasattr(ee, 'knl'):
        for i in range(1, len(ee.knl), 2):
            ee_ref.knl[i] = -(ee_ref.knl[i]._expr or ee_ref.knl[i]._value)

    if hasattr(ee, 'ksl'):
        for i in range(0, len(ee.ksl), 2):
            ee_ref.ksl[i] = -(ee_ref.ksl[i]._expr or ee_ref.ksl[i]._value)

    _exchange_fields('edge_entry_model', 'edge_exit_model')
    _exchange_fields('edge_entry_angle', 'edge_exit_angle')
    _exchange_fields('edge_entry_angle_fdown', 'edge_exit_angle_fdown')
    _exchange_fields('edge_entry_fint', 'edge_exit_fint')
    _exchange_fields('edge_entry_hgap', 'edge_exit_hgap')

def _resolve_lines_in_components(components, env):

    components = list(components) # Make a copy

    for ii, nn in enumerate(components):
        if (isinstance(nn, Place) and isinstance(nn.name, str)
                and nn.name in env.lines):
            nn.name = env.lines[nn.name]
        if isinstance(nn, str) and nn in env.lines:
            components[ii] = env.lines[nn]

    return components


import xobjects as xo

from ..general import _pkg_root
from ..random import RandomUniformAccurate, RandomExponential
from .elements import (
    SynchrotronRadiationRecord, Quadrupole, Sextupole,
    Octupole, Bend, Multipole, DipoleEdge, RBend, MultipoleEdge, Marker
)
from ..base_element import BeamElement

ID_RADIATION_FROM_PARENT = 10

xo.context_default.kernels.clear()

_common_xofields = {
    'radiation_flag': xo.Field(xo.Int64, default=ID_RADIATION_FROM_PARENT),
    'delta_taper': xo.Float64,
    'weight': xo.Float64,
}

def _slice_copy(self, **kwargs):
    out = BeamElement.copy(self, **kwargs)
    out._parent = None
    out.parent_name = self.parent_name
    return out


class ThinSliceQuadrupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Quadrupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_quadrupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        knl = self._parent.knl.copy() * self.weight
        ksl = self._parent.ksl.copy() * self.weight

        knl[1] += self._parent.k1 * self._parent.length * self.weight
        ksl[1] += self._parent.k1s * self._parent.length * self.weight

        length = self._parent.length * self.weight

        if self.radiation_flag == ID_RADIATION_FROM_PARENT:
            radiation_flag = self._parent.radiation_flag
        else:
            radiation_flag = self.radiation_flag

        out = Multipole(knl=knl, ksl=ksl, length=length,
                        hxl=0,
                        radiation_flag=radiation_flag,
                        delta_taper=self.delta_taper,
                        shift_x=self._parent.shift_x,
                        shift_y=self._parent.shift_y,
                        shift_s=self._parent.shift_s,
                        rot_s_rad=self._parent.rot_s_rad,
                        _buffer=self._buffer)
        return out


class ThinSliceSextupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Sextupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_sextupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        knl = self._parent.knl.copy() * self.weight
        ksl = self._parent.ksl.copy() * self.weight

        knl[2] += self._parent.k2 * self._parent.length * self.weight
        ksl[2] += self._parent.k2s * self._parent.length * self.weight

        length = self._parent.length * self.weight

        if self.radiation_flag == ID_RADIATION_FROM_PARENT:
            radiation_flag = self._parent.radiation_flag
        else:
            radiation_flag = self.radiation_flag

        out = Multipole(knl=knl, ksl=ksl, length=length,
                        hxl=0,
                        radiation_flag=radiation_flag,
                        delta_taper=self.delta_taper,
                        shift_x=self._parent.shift_x,
                        shift_y=self._parent.shift_y,
                        shift_s=self._parent.shift_s,
                        rot_s_rad=self._parent.rot_s_rad,
                        _buffer=self._buffer)
        return out


class ThinSliceOctupole(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Octupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_octupole.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        knl = self._parent.knl.copy() * self.weight
        ksl = self._parent.ksl.copy() * self.weight

        knl[3] += self._parent.k3 * self._parent.length * self.weight
        ksl[3] += self._parent.k3s * self._parent.length * self.weight

        length = self._parent.length * self.weight

        if self.radiation_flag == ID_RADIATION_FROM_PARENT:
            radiation_flag = self._parent.radiation_flag
        else:
            radiation_flag = self.radiation_flag

        out = Multipole(knl=knl, ksl=ksl, length=length,
                        hxl=0,
                        radiation_flag=radiation_flag,
                        delta_taper=self.delta_taper,
                        shift_x=self._parent.shift_x,
                        shift_y=self._parent.shift_y,
                        shift_s=self._parent.shift_s,
                        rot_s_rad=self._parent.rot_s_rad,
                        _buffer=self._buffer)
        return out


class ThinSliceBend(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(Bend), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_bend.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        knl = self._parent.knl.copy() * self.weight
        ksl = self._parent.ksl.copy() * self.weight

        knl[0] += self._parent.k0 * self._parent.length * self.weight
        knl[1] += self._parent.k1 * self._parent.length * self.weight

        length = self._parent.length * self.weight

        if self.radiation_flag == ID_RADIATION_FROM_PARENT:
            radiation_flag = self._parent.radiation_flag
        else:
            radiation_flag = self.radiation_flag

        out = Multipole(knl=knl, ksl=ksl, length=length,
                        hxl=self._parent.h * length,
                        radiation_flag=radiation_flag,
                        delta_taper=self.delta_taper,
                        shift_x=self._parent.shift_x,
                        shift_y=self._parent.shift_y,
                        shift_s=self._parent.shift_s,
                        rot_s_rad=self._parent.rot_s_rad,
                        _buffer=self._buffer)
        return out


class ThinSliceBendEntry(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Bend), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_dipole_edge_nonlinear.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_bend_entry.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        if self._parent.edge_entry_active:
            return DipoleEdge(
                k=self._parent.k0,
                e1=self._parent.edge_entry_angle,
                e1_fd=self._parent.edge_entry_angle_fdown,
                hgap=self._parent.edge_entry_hgap,
                fint=self._parent.edge_entry_fint,
                model=self._parent.edge_entry_model,
                side='entry',
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceBendExit(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Bend), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_dipole_edge_nonlinear.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_bend_exit.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        if self._parent.edge_exit_active:
            return DipoleEdge(
                k=self._parent.k0,
                e1=self._parent.edge_exit_angle,
                e1_fd=self._parent.edge_exit_angle_fdown,
                hgap=self._parent.edge_exit_hgap,
                fint=self._parent.edge_exit_fint,
                model=self._parent.edge_exit_model,
                delta_taper=self.delta_taper,
                side='exit',
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceQuadrupoleEntry(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Quadrupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_quadrupole_entry.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        if self._parent.edge_entry_active:
            return MultipoleEdge(
                kn=[0, self._parent.k1],
                ks=[0, self._parent.k1s],
                is_exit=False,
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceQuadrupoleExit(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Quadrupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_quadrupole_exit.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        if self._parent.edge_exit_active:
            return MultipoleEdge(
                kn=[0, self._parent.k1],
                ks=[0, self._parent.k1s],
                is_exit=True,
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceSextupoleEntry(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Sextupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_sextupole_entry.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        if self._parent.edge_entry_active:
            return MultipoleEdge(
                kn=[0, 0, self._parent.k2],
                ks=[0, 0, self._parent.k2s],
                is_exit=False,
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceSextupoleExit(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Sextupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_sextupole_exit.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        if self._parent.edge_exit_active:
            return MultipoleEdge(
                kn=[0, 0, self._parent.k2],
                ks=[0, 0, self._parent.k2s],
                is_exit=True,
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceOctupoleEntry(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Octupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_octupole_entry.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        if self._parent.edge_entry_active:
            return MultipoleEdge(
                kn=[0, 0, 0, self._parent.k3],
                ks=[0, 0, 0, self._parent.k3s],
                is_exit=False,
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceOctupoleExit(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(Octupole), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_octupole_exit.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        if self._parent.edge_exit_active:
            return MultipoleEdge(
                kn=[0, 0, 0, self._parent.k3],
                ks=[0, 0, 0, self._parent.k3s],
                is_exit=True,
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceRBend(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    _depends_on = [RandomUniformAccurate, RandomExponential]
    _internal_record_class = SynchrotronRadiationRecord
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = True

    _xofields = {'_parent': xo.Ref(RBend), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_rbend.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):
        knl = self._parent.knl.copy() * self.weight
        ksl = self._parent.ksl.copy() * self.weight

        knl[0] += self._parent.k0 * self._parent.length * self.weight
        knl[1] += self._parent.k1 * self._parent.length * self.weight

        length = self._parent.length * self.weight

        out = Multipole(knl=knl, ksl=ksl, length=length,
                        hxl=self._parent.h * length,
                        shift_x=self._parent.shift_x,
                        shift_y=self._parent.shift_y,
                        shift_s=self._parent.shift_s,
                        rot_s_rad=self._parent.rot_s_rad,
                        _buffer=self._buffer)
        return out


class ThinSliceRBendEntry(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(RBend), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_dipole_edge_nonlinear.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_rbend_entry.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        if self._parent.edge_entry_active:
            return DipoleEdge(
                k=self._parent.k0,
                e1=self._parent.edge_entry_angle + self._parent.angle / 2,
                e1_fd=self._parent.edge_entry_angle_fdown,
                hgap=self._parent.edge_entry_hgap,
                fint=self._parent.edge_entry_fint,
                model=self._parent.edge_entry_model,
                side='entry',
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)


class ThinSliceRBendExit(BeamElement):
    allow_rot_and_shift = False
    rot_and_shift_from_parent = True
    _skip_in_to_dict = ['_parent']
    has_backtrack = True
    _force_moveable = True
    _inherit_strengths = False

    _xofields = {'_parent': xo.Ref(RBend), **_common_xofields}

    _extra_c_sources = [
        _pkg_root.joinpath('headers/constants.h'),
        _pkg_root.joinpath('beam_elements/elements_src/track_dipole_edge_nonlinear.h'),
        _pkg_root.joinpath('beam_elements/elements_src/thin_slice_rbend_exit.h')]

    copy = _slice_copy

    def to_dict(self, **kwargs):
        dct = BeamElement.to_dict(self, **kwargs)
        dct['parent_name'] = self.parent_name
        return dct

    @classmethod
    def from_dict(cls, dct, **kwargs):
        obj = super().from_dict(dct, **kwargs)
        obj.parent_name = dct['parent_name']
        return obj

    def get_equivalent_element(self):

        if self._parent.edge_exit_active:
            return DipoleEdge(
                k=self._parent.k0,
                e1=self._parent.edge_exit_angle + self._parent.angle / 2,
                e1_fd=self._parent.edge_exit_angle_fdown,
                hgap=self._parent.edge_exit_hgap,
                fint=self._parent.edge_exit_fint,
                model=self._parent.edge_exit_model,
                side='exit',
                shift_x=self._parent.shift_x,
                shift_y=self._parent.shift_y,
                shift_s=self._parent.shift_s,
                rot_s_rad=self._parent.rot_s_rad,
                _buffer=self._buffer
            )
        else:
            return Marker(_buffer=self._buffer)




import numpy as np
from scipy.constants import c as clight

import xtrack as xt
import xpart as xp
import xobjects as xo
from functools import partial

line = xt.Line.from_json(
    # '../../test_data/hllhc14_no_errors_with_coupling_knobs/line_b1.json')
    '../../test_data/hllhc15_thick/lhc_thick_with_knobs.json')
# line.cycle('ip1')
line.build_tracker()
line.vv['vrf400'] = 16

for vv in line.vars.get_table().rows[
    'on_x.*|on_sep.*|on_crab.*|on_alice|on_lhcb|corr_.*'].name:
    line.vars[vv] = 0

tw = line.twiss()

df_hz = -50
h_rf = 35640
f_rev = 1/tw.T_rev0
df_rev = df_hz / h_rf
eta = tw.slip_factor

f_rf0 = 1/tw.T_rev0 * h_rf

f_rf = f_rf0 + df_hz
line.vars['f_rf'] = f_rf
tt = line.get_table()
for nn in tt.rows[tt.element_type=='Cavity'].name:
    line.element_refs[nn].absolute_time = 1 # Need property
    line.element_refs[nn].frequency = line.vars['f_rf']

tw1 = line.twiss(search_for_t_rev=True)

f_rev_expected = f_rf / h_rf

assert np.isclose(f_rev_expected, 1/tw1.T_rev, atol=1e-5, rtol=0)
assert np.allclose(tw1.delta, tw1.delta[0], atol=1e-5, rtol=0) # Check that it is flat
delta_expected = -df_rev / f_rev / eta
assert np.allclose(tw1.delta, delta_expected, atol=2e-6, rtol=0)
tw_off_mom = line.twiss(method='4d', delta0=tw1.delta[0])
assert np.allclose(tw1.x, tw_off_mom.x, atol=1e-5, rtol=0)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw1.s, tw1.x*1e3, label='x')
plt.ylabel('x [mm]')
plt.grid()
ax2 = plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw1.s, tw1.delta*1e3, label='delta')
plt.xlabel('s [m]')
plt.ylabel(r'$\delta$ [$10^{-3}$]')
plt.ylim(-0.5, 0.5)
plt.grid()

plt.suptitle(r'$\Delta f_{\mathrm{rf}}$ = ' +f'{df_hz} Hz, '
             f'expexted $\delta$ = {delta_expected*1e3:.2f}e-3')

plt.show()


import xtrack as xt
import xobjects as xo

env = xt.Environment()

line = env.new_line(name='line', components=[
    env.new('qd', xt.Quadrupole, k1=0.1, length=0.1),
])
line_thick = line.copy(shallow=True)
line.slice_thick_elements(slicing_strategies=[xt.Strategy(slicing=xt.Teapot(3))])

p0 = xt.Particles(kinetic_energy0=100e6, mass0=xt.PROTON_MASS_EV,
                  x = 1e-3, px=1e-3, y=1e-3, py=1e-3, zeta=1e-3, delta=1e-3)

p_ref_no_fringe = p0.copy()
p_slice_no_fringe = p0.copy()

line_thick.track(p_ref_no_fringe)
line.track(p_slice_no_fringe)

line.configure_quadrupole_model(edge='full')
p_ref_fringe = p0.copy()
p_slice_fringe = p0.copy()

line_thick.track(p_ref_fringe)
line.track(p_slice_fringe)

diff_no_fringe_x = p_ref_no_fringe.x - p_slice_no_fringe.x
diff_fringe_x = p_ref_fringe.x - p_slice_fringe.x

xo.assert_allclose(diff_no_fringe_x, diff_fringe_x, rtol=0, atol=1e-16)



# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import xtrack as xt
import xdeps as xd

## Generate a simple line
line = xt.Line(
    elements=[xt.Drift(length=2.),
              xt.Multipole(knl=[0, 1.], ksl=[0,0]),
              xt.Drift(length=1.),
              xt.Multipole(knl=[0, -1.], ksl=[0,0])],
    element_names=['drift_0', 'quad_0', 'drift_1', 'quad_1'])

# manager=xd.Manager()
# rele = manager.ref(line.element_dict, 'ele')

# knobs = {}
# rknobs = manager.ref(knobs, 'knobs')

# # We control a scalar
# rknobs['ldrifts'] = 3
# rele['drift_0'].length = rknobs['ldrifts']
# rele['drift_1'].length = rknobs['ldrifts']
# rknobs['ldrifts'] = 5
# assert line.elements[0].length == 5
# assert line.elements[2].length == 5

# rknobs['quadcorr'] = 0.1
# rele['quad_0'].knl[1] = 1. + rknobs['quadcorr']
# rele['quad_1'].knl[1] = -1. - rknobs['quadcorr']

# assert line.elements[1].knl[1] == 1.1
# assert line.elements[3].knl[1] == -1.1

# rknobs['quadcorr'] = 0.2
# assert line.elements[1].knl[1] == 1.2
# assert line.elements[3].knl[1] == -1.2


import xtrack as xt

import detuning

nemitt_x = 1e-6
nemitt_y = 1e-6

line = xt.Line.from_json(
    '../../test_data/hllhc15_noerrors_nobb/line_w_knobs_and_particle.json')
line.particle_ref = xt.Particles(mass0=xt.PROTON_MASS_EV, p0c=7e12)
line.build_tracker()

for sec in [12, 23, 34, 45, 56, 67, 78, 81]:
    for tt in ['f', 'd']:
        line.vars[f"ko{tt}.a{sec}b1"] = 18

axx, axy, ayx, ayy = detuning.get_amplitude_detuning(line)

print(f"axx = {1.e-6*axx:.3f} um^-1")
print(f"axy = {1.e-6*axy:.3f} um^-1")
print(f"ayx = {1.e-6*ayx:.3f} um^-1")
print(f"ayy = {1.e-6*ayy:.3f} um^-1")
# expecting
# axx = 1.070 um^-1
# axy = -0.573 um^-1
# ayx = -0.514 um^-1
# ayy = 1.073 um^-1

order=3
chromaticity = detuning.get_nonlinear_chromaticity(line, order=order)

print()
for ii in range(2, order+1):
    print("Q" + ii *"'" + (order - ii)*" " + f"x = {chromaticity.qx_derivatives[ii]:.3e}")
for ii in range(2, order+1):
    print("Q" + ii *"'" + (order - ii)*" " + f"y = {chromaticity.qx_derivatives[ii]:.3e}")



import numpy as np
from cpymad.madx import Madx
from pymadng import MAD
import xtrack as xt
import uuid

madx1 = Madx()
madx1.call("../../test_data/hllhc15_thick/lhc.seq")
madx1.call("../../test_data/hllhc15_thick/hllhc_sequence.madx")
madx1.call("../../test_data/hllhc15_thick/opt_round_150_1500.madx")
madx1.beam(particle='proton', energy=7000e9)
madx1.use("lhcb1")
madx1.input('save, sequence=lhcb1, file="formadn.seq";')

mng1 = MAD()
mng1.MADX.load('"formadn.seq"', f"'mad1.madng'")
mng1["lhcb1"] = mng1.MADX.lhcb1
mng1.lhcb1.beam = mng1.beam()
mng1["mytwtable", 'mytwflow'] = mng1.twiss(
    sequence=mng1.lhcb1, method=4, mapdef=2, implicit=True, nslice=3, save="'atbody'")


line = xt.Line.from_madx_sequence(sequence=madx1.sequence.lhcb1,
                                    deferred_expressions=True)
line.particle_ref = xt.Particles(p0c=7000e9, mass0=xt.PROTON_MASS_EV)

tw = line.twiss(method='4d')

mng2 = line.to_madng(sequence_name='lhcb1')

mng2["lhcb1"] = mng2.MADX.lhcb1
mng2.lhcb1.beam = mng2.beam(particle="'proton'", energy=7000)
mng2["mytwtable", 'mytwflow'] = mng2.twiss(
    sequence=mng2.lhcb1, method=4, mapdef=2, implicit=True, nslice=3, save="'atbody'")

print(mng2["mytwtable"].mu1[-1])
assert np.isclose(mng2["mytwtable"].mu1[-1][0], 62.31, atol=1e-6, rtol=0)


mng2.send('''
    local track in MAD  -- like "from MAD import track"
    local mytrktable, mytrkflow = MAD.track{sequence=MADX.lhcb1, method=4,
                                            mapdef=4, nslice=3}

        -- print(MAD.typeid.is_damap(mytrkflow[1]))

    local normal in MAD.gphys  -- like "from MAD.gphys import normal"
    local my_norm_for = normal(mytrkflow[1]):analyse('anh') -- anh stands for anharmonicity

    local nf = my_norm_for
    py:send({
            nf:q1{1}, -- qx from the normal form (fractional part)
            nf:q2{1}, -- qy
            nf:dq1{1}, -- dqx / d delta
            nf:dq2{1}, -- dqy / d delta
            nf:dq1{2}, -- d2 qx / d delta2
            nf:dq2{2}, -- d2 qy / d delta2
            nf:anhx{1, 0}, -- dqx / djx
            nf:anhy{0, 1}, -- dqy / djy
            nf:anhx{0, 1}, -- dqx / djy
            nf:anhy{1, 0}, -- dqy / djx
            })
''')

out = mng2.recv()

madx2 = Madx()
madx2.call("xsuite_to_mad.madx")
madx2.beam(particle='proton', energy=7000)
madx2.use("lhcb1")
tmx2 = madx2.twiss()

dp = 1e-5
twp = line.twiss(method='4d', delta0=dp)
twm = line.twiss(method='4d', delta0=-dp)

ddqx = (twp.dqx - twm.dqx) / (2*dp)





import xtrack as xt

env = xt.Environment()

env.new('q1', 'Quadrupole', length=2.0)
components=[]

for ii in range(1000):
    components.extend([
        env.new(f'c{ii}q1', 'Quadrupole', length=2.0, anchor='start', at=1. + 50 * ii),
        env.new(f'c{ii}q2', 'q1', anchor='start', at=10., from_=f'c{ii}q1@end'),
        env.new(f'c{ii}s2', 'Sextupole', length=0.1, anchor='end', at=-1., from_=f'c{ii}q2@start'),
        env.new(f'c{ii}q3', 'Quadrupole', length=2.0, at=20.+ 50 * ii),
        env.new(f'c{ii}q4', f'c{ii}q3', anchor='start', at=f'c{ii}q3@end'),
        env.new(f'c{ii}q5', f'c{ii}q3'),
        env.new(f'c{ii}m2', 'Marker', at=f'c{ii}q2@start'),
        env.new(f'c{ii}m2_0', 'Marker', at=f'c{ii}m2@start'),
        env.new(f'c{ii}m2_1', 'Marker', at=f'c{ii}m2@end'),
        env.new(f'c{ii}m2_1_0', 'Marker', at=f'c{ii}m2_1@start'),
        env.new(f'c{ii}m1', 'Marker', at=f'c{ii}q1@start'),
        env.new(f'c{ii}m4', 'Marker', at=f'c{ii}q4@start'),
        env.new(f'c{ii}m3', 'Marker', at=f'c{ii}q3@end'),
    ])

print('Components built')

import time
t0 = time.time()
line = env.new_line(components=components)
t1 = time.time()
print(f'Line built in {t1-t0} s')

# line.get_table().show(cols=['name', 's_start', 's_end', 's_center'])


from .random_generators import (RandomUniform, RandomExponential, RandomNormal, RandomRutherford,
                                RandomUniformAccurate)

rng_classes = (RandomUniform, RandomExponential, RandomNormal, RandomRutherford,
                RandomUniformAccurate)



import itertools
import json
import os
from typing import Literal

import numpy as np
import xtrack as xt
import xobjects as xo

if not os.path.exists('json.lua'):
    os.system('curl -OL https://raw.githubusercontent.com/rxi/json.lua/master/json.lua')

MADNG_EXEC = '/opt/madng/madng'
MODE: Literal['save', 'discard'] = 'save'
XT_ELEMENTS = {
    'quadrupole': xt.Quadrupole,
    'sextupole': xt.Sextupole,
    'octupole': xt.Octupole,
}
WHICH_KN = {
    'quadrupole': 'k1',
    'sextupole': 'k2',
    'octupole': 'k3',
}

def load_json(file_path):
    with open(file_path) as f:
        return json.load(f)

def coord_list(max_value):
    return list(np.linspace(-max_value, max_value, 7))

values_for_dims = [
    coord_list(1e-3),  # x
    coord_list(1e-5),  # px
    coord_list(1e-3),  # y
    coord_list(1e-5),  # py
    coord_list(1e-2),  # zeta
    coord_list(1e-3),  # delta
]
all_coords = np.array(list(itertools.product(*values_for_dims))).T

p = xt.Particles(
    x=all_coords[0],
    px=all_coords[1],
    y=all_coords[2],
    py=all_coords[3],
    zeta=all_coords[4],
    delta=all_coords[5],
    energy0=2e9,  # 2 GeV
    mass0=xt.ELECTRON_MASS_EV,
    q0=1,
    beta0=1,
)
p.to_json('initial_particles.json')

length = 0.01
knl = 1

for element in ('quadrupole', 'sextupole', 'octupole'):
    p_xt = {}

    for has_fringe in (True, False):
        _p = p.copy()
        line = xt.Line(
            elements=[
                XT_ELEMENTS[element](
                    length=length,
                    edge_entry_active=has_fringe,
                    edge_exit_active=has_fringe,
                    **{WHICH_KN[element]: knl / length},
                ),
            ],
            element_names=['q'],
        )
        line.build_tracker()
        line.track(_p)
        _p.sort()

        p_xt[has_fringe] = _p

    coords_xtrack_no_fringe = np.array([
        p_xt[0].x,
        p_xt[0].px,
        p_xt[0].y,
        p_xt[0].py,
        p_xt[0].zeta / p_xt[0].beta0,
        p_xt[0].ptau]
    )
    coords_xtrack_fringe = np.array([
        p_xt[1].x,
        p_xt[1].px,
        p_xt[1].y,
        p_xt[1].py,
        p_xt[1].zeta / p_xt[1].beta0,
        p_xt[1].ptau],
    )

    with open('config.json', 'w') as config_file:
        json.dump({
            'element': element,
            'param': WHICH_KN[element],
        }, config_file)

    if os.system(f'{MADNG_EXEC} fringe.lua'):
        raise RuntimeError("MAD-NG script failed, make sure to point "
                           "`MADNG_EXEC` to your madng executable path")

    p_fringe = load_json('track_fringe.json')
    p_no_fringe = load_json('track_no_fringe.json')

    coords_madng_fringe = np.array([p_fringe[coord] for coord in ('x', 'px', 'y', 'py', 't', 'pt')])
    coords_madng_no_fringe = np.array([p_no_fringe[coord] for coord in ('x', 'px', 'y', 'py', 't', 'pt')])

    print("Comparing mad-ng with and without fringes:")
    max_ng_errors = np.max(abs(coords_madng_fringe - coords_madng_no_fringe), axis=1)
    print(f"Max error: {', '.join(str(x) for x in max_ng_errors)}")

    print()

    print("Comparing mad-ng with no fringes and xtrack:")
    max_errors = np.max(abs(coords_madng_no_fringe - coords_xtrack_no_fringe), axis=1)
    print(f"Max error: {', '.join(str(x) for x in max_errors)}")

    avg_errors = np.mean(abs(coords_madng_no_fringe - coords_xtrack_no_fringe), axis=1)
    print(f"Avg error: {', '.join(str(x) for x in avg_errors)}")


    print()

    print("Comparing mad-ng and xtrack:")
    max_errors = np.max(abs(coords_madng_fringe - coords_xtrack_fringe), axis=1)
    print(f"Max error: {', '.join(str(x) for x in max_errors)}")

    avg_errors = np.mean(abs(coords_madng_fringe - coords_xtrack_fringe), axis=1)
    print(f"Avg error: {', '.join(str(x) for x in avg_errors)}")

    print()

    print("Particles after tracking:")
    print(f"Xtrack with fringes: {coords_xtrack_fringe}")
    print(f"Xtrack without fringes: {coords_xtrack_no_fringe}")
    print(f"MAD-NG with fringes: {coords_madng_fringe}")
    print(f"MAD-NG without fringes: {coords_madng_no_fringe}")

    print("Effect of the fringes")
    print(f"Xtrack with fringes - Xtrack without fringes: {coords_xtrack_fringe - coords_xtrack_no_fringe}")
    print(f"MAD-NG with fringes - MAD-NG without fringes: {coords_madng_fringe - coords_madng_no_fringe}")

    fringe_effect_madng = coords_madng_fringe - coords_madng_no_fringe
    if MODE == 'save':
        with open(f'{element}_fringe.json', 'w') as f:
            json.dump(fringe_effect_madng, f, indent=2, cls=xo.JEncoder)

        fringe_effect_madng = load_json(f'{element}_fringe.json')

    fringe_effect_xtrack = coords_xtrack_fringe - coords_xtrack_no_fringe

    xo.assert_allclose(fringe_effect_madng[0], fringe_effect_xtrack[0], atol=1e-16, rtol=1.3e-2)  # x
    xo.assert_allclose(fringe_effect_madng[1], fringe_effect_xtrack[1], atol=1e-16, rtol=1.3e-2)  # px
    xo.assert_allclose(fringe_effect_madng[2], fringe_effect_xtrack[2], atol=1e-16, rtol=1e-2)  # y
    xo.assert_allclose(fringe_effect_madng[3], fringe_effect_xtrack[3], atol=1e-16, rtol=1.1e-2)  # py
    xo.assert_allclose(fringe_effect_madng[4], fringe_effect_xtrack[4], atol=1e-15, rtol=1e-3)  # t
    xo.assert_allclose(fringe_effect_madng[5], fringe_effect_xtrack[5], atol=0, rtol=0)  # pt



import json

from cpymad.madx import Madx
import xpart as xp
import xtrack as xt
import xobjects as xo

mad = Madx()
mad.call('psb_injection.seq')
mad.use('psb')
mad.twiss()

line = xt.Line.from_madx_sequence(mad.sequence['psb'], deferred_expressions=True)
line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1.,
                                 gamma0=mad.sequence['psb'].beam.gamma)

with open('line_and_particle.json', 'w') as fid:
    json.dump(line.to_dict(), fid, cls=xo.JEncoder)



import xtrack as xt
import xobjects as xo
import numpy as np

# Build line with half a cell
half_cell = xt.Line(
    elements={
        'start_cell': xt.Marker(),
        'drift0': xt.Drift(length=1.),
        'qf1': xt.Quadrupole(k1=0.027/2, length=1.),
        'drift1_1': xt.Drift(length=1),
        'bend1': xt.Bend(k0=3e-4, h=3e-4, length=45.),
        'drift1_2': xt.Drift(length=1.),
        'qd1': xt.Quadrupole(k1=-0.0271/2, length=1.),
        'drift2': xt.Drift(length=1),
        'mid_cell': xt.Marker(),
    }
)
half_cell.particle_ref = xt.Particles(p0c=2e9)

# Add observation points every 1 m (to see betas inside bends)
half_cell.discard_tracker()
s_cut = np.arange(0, half_cell.get_length(), 1.)
half_cell.cut_at_s(s_cut)

tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', # <--- periodic-symmetric boundary
                                 strengths=True # to get the strengths in table
                                )

cell = xt.Line(
    elements={
        'start_cell': xt.Marker(),
        'drift0': xt.Drift(length=1.),
        'qf1':    xt.Quadrupole(k1=0.027/2, length=1.),
        'drift1': xt.Drift(length=1),
        'bend1':  xt.Bend(k0=3e-4, h=3e-4, length=45.),
        'drift2': xt.Drift(length=1.),
        'qd1':    xt.Quadrupole(k1=-0.0271/2, length=1.),
        'drift3': xt.Drift(length=1),
        'mid_cell': xt.Marker(),
        'drift4': xt.Replica('drift3'),
        'qd2':    xt.Replica('qd1'),
        'drift5': xt.Replica('drift2'),
        'bend2':  xt.Replica('bend1'),
        'drift6': xt.Replica('drift1'),
        'qf2':    xt.Replica('qf1'),
        'drift7': xt.Replica('drift0'),
        'end_cell': xt.Marker(),
    }
)
cell.particle_ref = xt.Particles(p0c=2e9)

# Add observation points every 1 m (to see betas inside bends)
cell.discard_tracker()
s_cut = np.arange(0, cell.get_length(), 1.)
cell.cut_at_s(s_cut)

tw_cell = cell.twiss4d(strengths=True)

xo.assert_allclose(tw_half_cell.betx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].betx, atol=0, rtol=1e-8)
xo.assert_allclose(tw_half_cell.bety[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].bety, atol=0, rtol=1e-8)
xo.assert_allclose(tw_half_cell.alfx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].alfx, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.alfy[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].alfy, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.dx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].dx, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.dpx[:-1], # remove '_end_point'
                     tw_cell.rows[:'mid_cell'].dpx, atol=1e-8, rtol=0)

xo.assert_allclose(tw_half_cell.ax_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].ax_chrom, atol=1e-5, rtol=0)
xo.assert_allclose(tw_half_cell.ay_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].ay_chrom, atol=1e-5, rtol=0)
xo.assert_allclose(tw_half_cell.bx_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].bx_chrom, atol=1e-5, rtol=0)
xo.assert_allclose(tw_half_cell.by_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].by_chrom, atol=1e-5, rtol=0)

xo.assert_allclose(tw_half_cell.qx, tw_cell.qx / 2, atol=1e-9, rtol=0)
xo.assert_allclose(tw_half_cell.qy, tw_cell.qy / 2, atol=1e-9, rtol=0)
xo.assert_allclose(tw_half_cell.dqx, tw_cell.dqx / 2, atol=1e-6, rtol=0)
xo.assert_allclose(tw_half_cell.dqy, tw_cell.dqy / 2, atol=1e-6, rtol=0)

for ll in [cell, half_cell]:
    ll.vars['kqf'] = 0.027/2
    ll.vars['kqd'] = -0.0271/2
    ll.element_refs['qf1'].k1 = ll.vars['kqf']
    ll.element_refs['qd1'].k1 = ll.vars['kqd']

tw_cell = cell.twiss4d(strengths=True)
tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', strengths=True)

xo.assert_allclose(tw_cell.mux[-1], 0.231, atol=1e-3, rtol=0)
xo.assert_allclose(tw_cell.muy[-1], 0.233, atol=1e-3, rtol=0)
xo.assert_allclose(tw_half_cell.mux[-1], 0.231 / 2, atol=1e-3, rtol=0)
xo.assert_allclose(tw_half_cell.muy[-1], 0.233 / 2, atol=1e-3, rtol=0)

opt_halfcell = half_cell.match(
    method='4d',
    start='start_cell', end='mid_cell',
    init='periodic_symmetric',
    targets=xt.TargetSet(mux=0.2501/2, muy=0.2502/2, at='mid_cell'),
    vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
)

tw_cell = cell.twiss4d(strengths=True)
tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', strengths=True)

xo.assert_allclose(tw_half_cell.mux[-1], 0.2501 / 2, atol=1e-3, rtol=0)
xo.assert_allclose(tw_half_cell.muy[-1], 0.2502 / 2, atol=1e-3, rtol=0)
xo.assert_allclose(tw_cell.mux[-1], 0.231, atol=1e-3, rtol=0) # unaffected
xo.assert_allclose(tw_cell.muy[-1], 0.233, atol=1e-3, rtol=0) # unaffected

opt_cell = cell.match(
    method='4d',
    start='start_cell', end='end_cell',
    init='periodic',
    targets=xt.TargetSet(mux=0.2501, muy=0.2502, at='end_cell'),
    vary=xt.VaryList(['kqf', 'kqd'], step=1e-5),
)

xo.assert_allclose(half_cell.vv['kqf'], cell.vv['kqf'], rtol=1e-9, atol=0)
xo.assert_allclose(half_cell.vv['kqd'], cell.vv['kqd'], rtol=1e-9, atol=0)

tw_cell = cell.twiss4d(strengths=True)
tw_half_cell = half_cell.twiss4d(init='periodic_symmetric', strengths=True)

xo.assert_allclose(tw_half_cell.mux[-1], 0.2501 / 2, atol=1e-3, rtol=0)
xo.assert_allclose(tw_half_cell.muy[-1], 0.2502 / 2, atol=1e-3, rtol=0)
xo.assert_allclose(tw_cell.mux[-1], 0.2501, atol=1e-3, rtol=0)
xo.assert_allclose(tw_cell.muy[-1], 0.2502, atol=1e-3, rtol=0)

xo.assert_allclose(tw_half_cell.betx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].betx, atol=0, rtol=1e-8)
xo.assert_allclose(tw_half_cell.bety[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].bety, atol=0, rtol=1e-8)
xo.assert_allclose(tw_half_cell.alfx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].alfx, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.alfy[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].alfy, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.dx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].dx, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.dpx[:-1], # remove '_end_point'
                     tw_cell.rows[:'mid_cell'].dpx, atol=1e-8, rtol=0)
xo.assert_allclose(tw_half_cell.ddx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].ddx, atol=1e-7, rtol=0)
xo.assert_allclose(tw_half_cell.ddpx[:-1], # remove '_end_point'
                   tw_cell.rows[:'mid_cell'].ddpx, atol=1e-8, rtol=0)

xo.assert_allclose(tw_half_cell.ax_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].ax_chrom, atol=1e-5, rtol=0)
xo.assert_allclose(tw_half_cell.ay_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].ay_chrom, atol=1e-5, rtol=0)
xo.assert_allclose(tw_half_cell.bx_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].bx_chrom, atol=1e-5, rtol=0)
xo.assert_allclose(tw_half_cell.by_chrom[:-1], # remove '_end_point'
                    tw_cell.rows[:'mid_cell'].by_chrom, atol=1e-5, rtol=0)

xo.assert_allclose(tw_half_cell.qx, tw_cell.qx / 2, atol=1e-9, rtol=0)
xo.assert_allclose(tw_half_cell.qy, tw_cell.qy / 2, atol=1e-9, rtol=0)
xo.assert_allclose(tw_half_cell.dqx, tw_cell.dqx / 2, atol=1e-6, rtol=0)
xo.assert_allclose(tw_half_cell.dqy, tw_cell.dqy / 2, atol=1e-6, rtol=0)

tw_off_mom_cell = cell.twiss4d(strengths=True, delta0=1e-3)
tw_off_mom_half_cell = half_cell.twiss4d(
    init='periodic_symmetric', strengths=True, delta0=1e-3)

xo.assert_allclose(tw_off_mom_half_cell.x[:-1],
                   tw_off_mom_cell.rows[:'mid_cell'].x, atol=1e-12, rtol=0)

import matplotlib.pyplot as plt
plt.close('all')

fig1 = plt.figure(1, figsize=(6.4*1.2, 4.8))
ax1 = fig1.add_subplot(211)
plt_cell = tw_cell.plot(ax=ax1)
plt_cell.move_legend(left=1.4, bottom=1)
ax1.set_title('Full cell')

ax2 = fig1.add_subplot(212, sharex=ax1, sharey=ax1)
plt_half_cell = tw_half_cell.plot(ax=ax2)
plt_half_cell.move_legend(left=1.4, bottom=1)
ax2.set_title('Half cell')

ax1.set_xlim(0, tw_cell.s[-1])
fig1.subplots_adjust(right=0.73, hspace=0.5)

fig2 = plt.figure(2, figsize=(6.4*1.2, 4.8))
ax1 = fig2.add_subplot(211)
plt_cell = tw_cell.plot('ddx', ax=ax1)
plt_cell.move_legend(left=1.4, bottom=1)
plt_cell.ax.set_title('Full cell')

ax2 = fig2.add_subplot(212, sharex=ax1, sharey=ax1)
plt_half_cell = tw_half_cell.plot('ddx', ax=ax2)
plt_half_cell.move_legend(left=1.4, bottom=1)
plt_half_cell.ax.set_title('Half cell')

ax1.set_xlim(0, tw_cell.s[-1])
fig2.subplots_adjust(right=0.73, hspace=0.5)



plt.show()
